import os
import re
import shutil
import asyncio

from natsort import natsorted
from hachoir.parser import createParser
from hachoir.metadata import extractMetadata
from telethon.tl.types import InputMediaPhoto, InputMediaDocument
from telethon.errors import FloodWaitError
from instaloader import (
    Instaloader, Post, Profile, NodeIterator,
    TwoFactorAuthRequiredException, InvalidArgumentException, BadCredentialsException,
    ConnectionException, LoginRequiredException
)

from ..core.thumb_manage import get_thumbnail

# Thanks TO https://github.com/UsergeTeam/Userge/blob/alpha/userge/plugins/misc/instadl.py

# some helpers
def get_caption(post: Post) -> str:
    """ adds link to profile for tagged users """
    caption = post.caption
    replace = '<a href="https://instagram.com/{}/">{}</a>'
    for mention in post.caption_mentions:
        men = '@' + mention
        val = replace.format(mention, men)
        caption = caption.replace(men, val)
    header = f'♥️<code>{post.likes}</code>  💬<code>{post.comments}</code>'
    if post.is_video:
        header += f'  👀`{post.video_view_count}`'
    caption = header + '\n\n' + (caption or '')
    return caption


async def upload_to_tg(message, dirname: str, post: Post, sender_id: int) -> None:  # pylint: disable=R0912
    """ uploads downloaded post from local to telegram servers """
    pto = ("tk.jpg", ".jpeg", ".png", ".bmp")
    vdo = (".mkv", ".mp4", ".webm")
    paths = []
    if post.typename == 'GraphSidecar':
        # upload media group
        captioned = False
        caption = ""
        media = []
        for path in natsorted(os.listdir(dirname)):
            ab_path = dirname + '/' + path
            paths.append(ab_path)
            if str(path).endswith(pto):
                if captioned:
                    media.append(ab_path)
                else:
                    media.append(ab_path)
                    caption = get_caption(post)[:1023]
                    caption += f"\n\n<a href='tg://user?id={sender_id}'>completed</a>"
                    captioned = True
            elif str(path).endswith(vdo):
                if captioned:
                    media.append(ab_path)
                else:
                    media.append(ab_path)
                    caption = get_caption(post)[:1023]
                    caption += f"\n\n<a href='tg://user?id={sender_id}'>completed.</a>"
                    captioned = True
        if media:
            await message.client.send_file(message.chat_id, media,caption=caption,parse_mode="html", reply_to=message.id)
            #await message.client.send_media_group(Config.LOG_CHANNEL_ID, media)

    if post.typename == 'GraphImage':
        # upload a photo
        for path in natsorted(os.listdir(dirname)):
            if str(path).endswith(pto):
                ab_path = dirname + '/' + path
                paths.append(ab_path)
                await message.client.send_file(
                    message.chat_id,
                    ab_path,
                    caption=get_caption(post)[:1023] + f"\n\n<a href='tg://user?id={sender_id}'>Done</a>\n#uploads\n",
                    parse_mode="html",
                    reply_to=message.id)

    if post.typename == 'GraphVideo':
        # upload a video
        for path in natsorted(os.listdir(dirname)):
            if str(path).endswith(vdo):
                ab_path = dirname + '/' + path
                paths.append(ab_path)
                thumb = await get_thumbnail(ab_path)

                await message.client.send_file(
                    entity=message.chat_id,
                    file=ab_path,
                    thumb=thumb,
                    caption=get_caption(post)[:1023] + f"\n\n<a href='tg://user?id={sender_id}'>completed.</a>",
                    parse_mode="html",
                    reply_to=message.id)
                if thumb is not None:
                    try:
                        os.remove(thumb)
                    except:
                        pass
    for del_p in paths:
        if os.path.lexists(del_p):
            os.remove(del_p)


# run some process in threads?

def download_post(client: Instaloader, post: Post) -> bool:
    """ Downloads content and returns True """
    client.download_post(post, post.owner_username)
    return True



def get_post(client: Instaloader, shortcode: str) -> Post:
    """ returns a post object """
    return Post.from_shortcode(client.context, shortcode)


def get_profile(client: Instaloader, username: str) -> Profile:
    """ returns profile """
    return Profile.from_username(client.context, username)



def get_profile_posts(profile: Profile) -> NodeIterator[Post]:
    """ returns a iterable Post object """
    return profile.get_posts()


# pylint: disable=R0914, R0912, R0915, R0911
async def _insta_post_downloader(message):
    """ download instagram post """
    omess = await message.get_reply_message()
    if omess is None:
        await message.reply("reply to a instagram link.")
        return
    message = await message.reply('`setting up configs. please don\'t flood.`')
    dirname = 'instadl_{target}'
    filename = '{target}\'s_post'
    insta = Instaloader(
        dirname_pattern=dirname,
        filename_pattern=filename,
        download_video_thumbnails=False,
        download_geotags=False,
        download_comments=False,
        save_metadata=False,
        compress_json=False
    )
    if False:
        # add auth code here
        pass
    else:
        await message.edit('login credentials not found.\n`[NOTE]`: '
                           '**private stuff will not be downloaded**')
        await asyncio.sleep(2)

    p = r'^https:\/\/www\.instagram\.com\/(p|tv|reel)\/([A-Za-z0-9\-_]*)\/(\?igshid=[a-zA-Z0-9]*)?$'
    match = re.search(p, omess.raw_text)
    print(omess.raw_text)
    if False:
        # have plans here
        pass
    elif match:
        dtypes = {
            'p': 'POST',
            'tv': 'IGTV',
            'reel': 'REELS'
        }
        d_t = dtypes.get(match.group(1))
        if not d_t:
            await message.edit('unsupported format')
            return
        sent = await message.edit(f'`fetching {d_t} content.`')
        shortcode = match.group(2)
        post = get_post(insta, shortcode)
        try:
            download_post(insta, post)
            await upload_to_tg(message, dirname.format(target=post.owner_username), post, sender_id=omess.sender_id)
        except (KeyError, LoginRequiredException):
            await message.edit("post is private. unable to download.")
            return
        except FloodWaitError as f_w:
            await asyncio.sleep(f_w.seconds + 5)
            await upload_to_tg(message, dirname.format(target=post.owner_username), post, sender_id=omess.sender_id)
        finally:
            shutil.rmtree(dirname.format(target=post.owner_username), ignore_errors=True)
        
    else:
        await message.edit('`this is an invaild instagram link.`')


# -*- coding: utf-8 -*-


import os,logging
from time import time
import asyncio as aio
from PIL import Image
from hachoir.parser import createParser
from hachoir.metadata import extractMetadata
import math

torlog = logging.getLogger(__name__)

async def gen_ss(filepath, ts, opfilepath=None):
    # todo check the error pipe and do processing 
    source = filepath
    destination = os.path.dirname(source)
    ss_name =  str(os.path.basename(source)) + "_" + str(round(time())) + ".jpg"
    ss_path = os.path.join(destination,ss_name)

    cmd = ["ffmpeg","-loglevel","error","-ss",str(ts),"-i",str(source),"-vframes","1","-q:v","2",str(ss_path)]

    subpr = await aio.create_subprocess_exec(
        *cmd,
        stdout=aio.subprocess.PIPE,
        stderr=aio.subprocess.PIPE
    )
    spipe, epipe = await subpr.communicate()
    epipe = epipe.decode().strip()
    spipe = spipe.decode().strip()
    torlog.info("Stdout Pipe :- {}".format(spipe))
    torlog.info("Error Pipe :- {}".format(epipe))

    return ss_path

async def resize_img(path,width=None,height=None):
    img = Image.open(path)
    wei,hei = img.size

    wei = width if width is not None else wei
    hei = height if height is not None else hei

    img.thumbnail((wei,hei))
    
    img.save(path,"JPEG")
    return path

async def split_file(path,max_size,force_docs=False):
    
    metadata = extractMetadata(createParser(path))

    if metadata.has("duration"):
        total_duration = metadata.get('duration').seconds

    metadata = metadata.exportDictionary()
    try:
        mime = metadata.get("Common").get("MIME type")
    except:
        mime = metadata.get("Metadata").get("MIME type")
    
    

    ftype = mime.split("/")[0]
    ftype = ftype.lower().strip()

    split_dir = os.path.join(os.path.dirname(path),str(time()))

    if not os.path.isdir(split_dir):
        os.makedirs(split_dir)
    
    if ftype == "video" and not force_docs:
        total_file_size = os.path.getsize(path)
        
        parts = math.ceil(total_file_size/max_size)
        #need this to be implemented to remove recursive file split calls
        #remove saftey margin
        #parts += 1
        torlog.info(f"Parts {parts}")

        minimum_duration = (total_duration / parts) 
        
        #casting to int cuz float Time Stamp can cause errors
        minimum_duration = int(minimum_duration)
        torlog.info(f"Min dur :- {minimum_duration} total {total_duration}")

        # END: proprietary
        start_time = 0
        end_time = minimum_duration


        base_name = os.path.basename(path)
        input_extension = base_name.split(".")[-1]
        
        i = 0
        flag = False
        
        while end_time <= total_duration:

            #file name generate
            parted_file_name = "{}_PART_{}.{}".format(str(base_name),str(i).zfill(5),str(input_extension))

            output_file = os.path.join(split_dir, parted_file_name)
            
            opfile = await cult_small_video(
                path,
                output_file,
                str(start_time),
                str(end_time)
            )
            torlog.info(f"Output file {opfile}")
            torlog.info(f"Start time {start_time}, End time {end_time}, Itr {i}")

            #adding offset of 3 seconds to ensure smooth playback 
            start_time = end_time - 3
            end_time = end_time + minimum_duration
            i = i + 1

            if (end_time > total_duration) and not flag:
                 end_time = total_duration
                 flag = True
            elif i+1 == parts:
                end_time = total_duration
                flag = True
            elif flag:
                break

    return split_dir

async def cult_small_video(video_file, out_put_file_name, start_time, end_time):
    file_genertor_command = [
        "ffmpeg",
        "-hide_banner",
        "-i",
        video_file,
        "-ss",
        start_time,
        "-to",
        end_time,
        "-async",
        "1",
        "-strict",
        "-2",
        "-c",
        "copy",
        out_put_file_name
    ]
    process = await aio.create_subprocess_exec(
        *file_genertor_command,
        # stdout must a pipe to be accessible as process.stdout
        stdout=aio.subprocess.PIPE,
        stderr=aio.subprocess.PIPE,
    )
    # Wait for the subprocess to finish
    stdout, stderr = await process.communicate()
    e_response = stderr.decode().strip()
    t_response = stdout.decode().strip()
    return out_put_file_name


# -*- coding: utf-8 -*-


import asyncio,shlex,os,logging,time
from typing import Union,Optional,List,Tuple

# ref https://info.nrao.edu/computing/guide/file-access-and-archiving/7zip/7z-7za-command-line-guide
torlog = logging.getLogger(__name__)

# TODO change the hard coded value of the size from here

async def cli_call(cmd: Union[str,List[str]]) -> Tuple[str,str]:
    torlog.info("Got cmd:- "+str(cmd))
    if isinstance(cmd,str):
        cmd = shlex.split(cmd)
    elif isinstance(cmd,(list,tuple)):
        pass
    else:
        return None,None

    torlog.info("Exc cmd:- "+str(cmd))
    
    process = await asyncio.create_subprocess_exec(
        *cmd,
        stderr=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE
    )

    stdout, stderr = await process.communicate()
    
    stdout = stdout.decode().strip()
    stderr = stderr.decode().strip()
    
    return stdout, stderr, process.returncode

async def split_in_zip(path,size=None):
    if os.path.exists(path):
        if os.path.isfile(path):
            fname = os.path.basename(path)
            bdir = os.path.dirname(path)
            bdir = os.path.join(bdir,str(time.time()).replace(".",""))
            if not os.path.exists(bdir):
                os.mkdir(bdir)

            if size is None:
                size = 1900
            else:
                size = int(size)
                size = int(size/(1024*1024)) - 10 #for safe
            cmd = f'7z a -tzip -mx=0 "{bdir}/{fname}.zip" "{path}" -v{size}m '

            _, err, rcode = await cli_call(cmd)
            
            if err:
                torlog.error(f"Error in zip split {err}")
                return None
            else:
                return bdir
                
        else:
            return None
    else:
        return None

async def add_to_zip(path, size = None, split = True):
    if os.path.exists(path):
        fname = os.path.basename(path)
        bdir = os.path.dirname(path)
        bdir = os.path.join(bdir,str(time.time()).replace(".",""))
        if not os.path.exists(bdir):
            os.mkdir(bdir)
        
        bdir = os.path.join(bdir,fname)
        if not os.path.exists(bdir):
            os.mkdir(bdir)
        
        if size is None:
            size = 1900
        else:
            size = int(size)
            size = int(size/(1024*1024)) - 10 #for safe

        total_size = get_size(path)
        if total_size > size and split:
            cmd = f'7z a -tzip -mx=0 "{bdir}/{fname}.zip" "{path}" -v{size}m'
        else:
            cmd = f'7z a -tzip -mx=0 "{bdir}/{fname}.zip" "{path}"'
    
        _, err, rcode = await cli_call(cmd)
        
        if err:
            torlog.error(f"Error in zip split {err}")
            return None
        else:
            return bdir
    else:
        return None

def get_size(start_path = '.'):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size/(1024*1024)

async def extract_archive(path, password=""):
    if os.path.exists(path):
        if os.path.isfile(path):
            if str(path).endswith((".zip", "7z", "tar", "gzip2", "iso", "wim", "rar", "tar.gz","tar.bz2")):
                # check userdata
                userpath = os.path.join(os.getcwd(), "userdata")
                if not os.path.exists(userpath):
                    os.mkdir(userpath)
                
                extpath = os.path.join(userpath, str(time.time()).replace(".",""))
                os.mkdir(extpath)
                
                extpath = os.path.join(extpath,os.path.basename(path))
                if not os.path.exists(extpath):
                    os.mkdir(extpath)

                if str(path).endswith(("tar","tar.gz","tar.bz2")):
                    cmd = f'tar -xvf "{path}" -C "{extpath}" --warning=none'
                else:
                    cmd = f'7z e -y "{path}" "-o{extpath}" "-p{password}"'
                
                out, err, rcode = await cli_call(cmd)
                
                if err:
                    if "Wrong password" in err:
                        return "Wrong Password"
                    else:
                        torlog.error(err)
                        torlog.error(out)
                        return False
                else:
                    return extpath
        else:
            # False means that the stuff can be upload but cant be extracted as its a dir
            return False
    else:
        # None means fetal error and cant be ignored
        return None 

#7z e -y {path} {ext_path}
#/setpassword jobid password


# -*- coding: utf-8 -*-

import os,subprocess,logging,re,time,json,traceback
from telethon.tl.types import KeyboardButtonUrl
from tgtk import SessionVars
import asyncio as aio
import aiohttp
from requests.utils import requote_uri
from ..core.getVars import get_val
from .. import upload_db, var_db
from telethon.tl.types import KeyboardButtonCallback
from ..core.status.upload import RCUploadTask 

torlog = logging.getLogger(__name__)


async def rclone_driver(path,message, user_msg, dl_task):
    # this driver will need to do this
    # get the default drive
    conf_path = await get_config()
    if conf_path is None:
        torlog.info("the configuration file was not found.")
        return None
    else:
        drive_name = get_val("DEF_RCLONE_DRIVE")
        rem_base = get_val("RCLONE_BASE_DIR")
        edtime = get_val("EDIT_SLEEP_SECS")

        
        ul_task = RCUploadTask(dl_task)
        try:
            return await rclone_upload(path,message,user_msg,drive_name,rem_base,edtime,conf_path, ul_task)
        except:
            await ul_task.set_inactive()
            torlog.exception("something went wrong. check logs for more infomation.")
            return 

# add user prompt here
async def rclone_upload(path,message,user_msg,dest_drive,dest_base,edit_time,conf_path, task):
    # this function will need a driver for him :o
    if not os.path.exists(path):
        torlog.info(f"returning none cause the path {path} not found.")
        await task.set_inactive(f"returning none cause the path {path} not found")
        return None
    omsg = user_msg
    await task.set_original_message(omsg)
    upload_db.register_upload(omsg.chat_id, omsg.id)
    data = "upcancel {} {} {}".format(omsg.chat_id,omsg.id,omsg.sender_id)
    buts = [KeyboardButtonCallback("cancel upload.",data.encode("UTF-8"))]
    
    msg = await message.reply("<b>uploading to configured drive.... will be updated soon.",parse_mode="html", buttons=buts)
    await task.set_message(msg)

    if os.path.isdir(path):
        # handle dirs
        new_dest_base = os.path.join(dest_base,os.path.basename(path))
        # buffer size needs more testing though #todo
        if get_val("RSTUFF"):
            rclone_copy_cmd = [get_val("RSTUFF"),'copy',f'--config={conf_path}',str(path),f'{dest_drive}:{new_dest_base}','-f','- *.!qB','--buffer-size=1M','-P']
        else:
            rclone_copy_cmd = ['rclone','copy',f'--config={conf_path}',str(path),f'{dest_drive}:{new_dest_base}','-f','- *.!qB','--buffer-size=1M','-P']

        # spawn a process # attempt 1 # test 2
        rclone_pr = subprocess.Popen(
            rclone_copy_cmd,
            stdout=subprocess.PIPE
        )
        rcres = await rclone_process_display(rclone_pr,edit_time,msg, message, omsg, task)
        
        if rcres is False:
            await message.edit(message.text + "\ncanceled rclone upload.")
            await msg.delete()
            rclone_pr.kill()
            task.cancel = True
            await task.set_inactive("canceled rclone upload.")
            return task
            

        torlog.info("upload complete")
        gid = await get_glink(dest_drive,dest_base,os.path.basename(path),conf_path)
        torlog.info(f"Upload folder id :- {gid}")
        
        folder_link = f"https://drive.google.com/folderview?id={gid[0]}"

        buttons = []
        buttons.append(
            [KeyboardButtonUrl("drive url",folder_link)]
        )
        gd_index = get_val("GD_INDEX_URL")
        if gd_index:
            index_link = "{}/{}/".format(gd_index.strip("/"), gid[1])
            index_link = requote_uri(index_link)
            torlog.info("index link "+str(index_link))
            buttons.append(
                [KeyboardButtonUrl("index url",index_link)]
            )


        txtmsg = "<a href='tg://user?id={}'>finished leech.</a>\nuploaded folder:<code>{}</code>\nto drive.".format(omsg.sender_id,os.path.basename(path))
        
        await omsg.reply(txtmsg,buttons=buttons,parse_mode="html")
        await msg.delete()


    else:
        new_dest_base = dest_base
        # buffer size needs more testing though #todo
        if get_val("RSTUFF"):
            rclone_copy_cmd = [get_val("RSTUFF"),'copy',f'--config={conf_path}',str(path),f'{dest_drive}:{new_dest_base}','-f','- *.!qB','--buffer-size=1M','-P']
        else:
            rclone_copy_cmd = ['rclone','copy',f'--config={conf_path}',str(path),f'{dest_drive}:{new_dest_base}','-f','- *.!qB','--buffer-size=1M','-P']

        # spawn a process # attempt 1 # test 2
        rclone_pr = subprocess.Popen(
            rclone_copy_cmd,
            stdout=subprocess.PIPE
        )
        rcres = await rclone_process_display(rclone_pr,edit_time,msg, message, omsg, task)
        
        if rcres is False:
            await message.edit(message.text + "\ncanceled rclone upload")
            await msg.delete()
            rclone_pr.kill()
            task.cancel = True
            await task.set_inactive("canceled rclone upload")
            return task

        torlog.info("upload complete")
        gid = await get_glink(dest_drive,dest_base,os.path.basename(path),conf_path,False)
        torlog.info(f"Upload folder id :- {gid}")

        buttons = []

        file_link = f"https://drive.google.com/file/d/{gid[0]}/view"
        buttons.append(
            [KeyboardButtonUrl("drive url",file_link)]
        )
        gd_index = get_val("GD_INDEX_URL")
        if gd_index:
            index_link = "{}/{}".format(gd_index.strip("/"), gid[1])
            index_link = requote_uri(index_link)
            torlog.info("index link "+str(index_link))
            buttons.append(
                [KeyboardButtonUrl("index url",index_link)]
            )

        txtmsg = "<a href='tg://user?id={}'>finished leech.</a>\nuploaded file:<code>{}</code>\nto drive.".format(omsg.sender_id,os.path.basename(path))

        
        await omsg.reply(txtmsg,buttons=buttons,parse_mode="html")
        await msg.delete()

    upload_db.deregister_upload(message.chat_id, message.id)
    await task.set_inactive()
    return task

async def rclone_process_display(process,edit_time,msg, omessage, cancelmsg, task):
    blank=0
    sleeps = False
    start = time.time()
    while True:
        
        data = process.stdout.readline().decode()
        data = data.strip()
        mat = re.findall("transferred:.*eta.*",data)
        
        if mat is not None:
            if len(mat) > 0:
                sleeps = True
                if time.time() - start > edit_time:
                    start = time.time()
                    
                    await task.refresh_info(data)
                    await task.update_message()

        if data == "":
            blank += 1
            if blank == 20:
                break
        else:
            blank = 0

        if sleeps:
            if upload_db.get_cancel_status(cancelmsg.chat_id, cancelmsg.id):
                return False
            
            sleeps=False
            await aio.sleep(2)
            process.stdout.flush()


async def get_glink(drive_name,drive_base,ent_name,conf_path,isdir=True):
    print("Ent - ",ent_name)
    ent_name = re.escape(ent_name)
    filter_path = os.path.join(os.getcwd(),str(time.time()).replace(".","")+".txt")
    with open(filter_path,"w",encoding="UTF-8") as file:
        file.write(f"+ {ent_name}\n")
        file.write(f"- *")

    if isdir:
        if get_val("RSTUFF"):
            get_id_cmd = [get_val("RSTUFF"), "lsjson", f'--config={conf_path}', f"{drive_name}:{drive_base}", "--dirs-only", "-f", f"+ {ent_name}/", "-f", "- *"]
        else:
            get_id_cmd = ["rclone", "lsjson", f'--config={conf_path}', f"{drive_name}:{drive_base}", "--dirs-only", "-f", f"+ {ent_name}/", "-f", "- *"]
        #get_id_cmd = ["rclone", "lsjson", f'--config={conf_path}', f"{drive_name}:{drive_base}", "--dirs-only", f"--filter-from={filter_path}"]
    else:
        if get_val("RSTUFF"):
            get_id_cmd = [get_val("RSTUFF"), "lsjson", f'--config={conf_path}', f"{drive_name}:{drive_base}", "--files-only", "-f", f"+ {ent_name}", "-f", "- *"]
        else:
            get_id_cmd = ["rclone", "lsjson", f'--config={conf_path}', f"{drive_name}:{drive_base}", "--files-only", "-f", f"+ {ent_name}", "-f", "- *"]
        #get_id_cmd = ["rclone", "lsjson", f'--config={conf_path}', f"{drive_name}:{drive_base}", "--files-only", f"--filter-from={filter_path}"]


    # piping only stdout
    process = await aio.create_subprocess_exec(
        *get_id_cmd,
        stdout=aio.subprocess.PIPE
    )

    stdout, _ = await process.communicate()
    stdout = stdout.decode().strip()
    
    if os.path.exists(filter_path):
        os.remove(filter_path)
    
    try:
        data = json.loads(stdout)
        id = data[0]["ID"]
        name = data[0]["Name"]
        return (id, name)
    except Exception:
        torlog.error("error occured while getting id: {} {}".format(traceback.format_exc(), stdout))

async def get_config():
    # this car requires to access the blob

    config = get_val("RCLONE_CONFIG")
    if config is not None:
        if isinstance(config,str):
            if os.path.exists(config):
                return config
    
    db = var_db
    _, blob = db.get_variable("RCLONE_CONFIG")
    
    if blob is not None:
        fpath = os.path.join(os.getcwd(),"rclone-config.conf")
        with open(fpath,"wb") as fi:
            fi.write(blob)
        return fpath
    
    return None

# probably hotfix for rclone ban



#!/usr/bin/env python
# -*- coding: utf-8 -*-
__version__ = "1.0.6"
__author__ = "reaitten"

import logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s %(threadName)s %(name)s %(message)s",
    handlers=[logging.StreamHandler(),logging.FileHandler("tk_log.txt")]
)

from tgtk.core.wserver import start_server
from .core.database_handle import tkupload, tkdb, tktorrents, userdb
from .core.varholdern import VarHolder
import time

logging.info("database created")
upload_db = tkupload()
var_db = tkdb()
tor_db = tktorrents()
user_db = userdb()


uptime = time.time()
to_del = []
SessionVars = VarHolder(var_db)


import os
try:
    from .ExecVars import ExecVars
except:
    class ExecVars:
        # Set true if its VPS (most likely not.)
        IS_VPS = False
        
        API_HASH = os.environ.get("API_HASH")
        API_ID = int(os.environ.get("APP_ID", 12345))
        BOT_TOKEN = os.environ.get("BOT_TOKEN", "")
        BASE_URL_OF_BOT = os.environ.get("BASE_URL_OF_BOT", "")

        # Edit the server port if you want to keep it default though.
        SERVPORT = 80

        # ALLOWED USERS [ids of user or supergroup] seperate by commas
        ALD_USR = set(int(x) for x in os.environ.get("AUTH_CHANNEL", "").split())
        OWNER_ID = os.environ.get("OWNER_ID", "")
        
        # Google Drive Index Link should include the base dir also See readme for more info
        GD_INDEX_URL = os.environ.get("GD_INDEX_URL", "False")

        # Time to wait before edit message
        EDIT_SLEEP_SECS = int(os.environ.get("EDIT_SLEEP_TIME_OUT", 15))

        # Telegram Upload Limit (in bytes)
        TG_UP_LIMIT = 1700000000

        # Should force evething uploaded into Document
        FORCE_DOCUMENTS = os.environ.get("FORCE_DOCUMENTS", "False")

        # Chracter to use as a completed progress 
        COMPLETED_STR = os.environ.get("COMPLETED_STR", "█")

        # Chracter to use as a incomplete progress
        REMAINING_STR = os.environ.get("REMAINING_STR", "░")

        # DB URI for access
        DB_URI = os.environ.get("DATABASE_URL", "")
        
        # UNCOMMENT THE BELOW LINE WHEN USING CONTAINER AND COMMENT THE UPPER LINE
        #DB_URI = "dbname=tortk user=postgres password=your-pass host=db port=5432"
        
        # The base direcory to which the files will be upload if using RCLONE
        RCLONE_BASE_DIR = os.environ.get("RCLONE_BASE_DIR", "/") 

        # This value will be considered only if Rclone is True - this may be defied now ;)
        # Cuz at least one needs to be Ture at a time either RCLONE or Leech.
        LEECH_ENABLED = os.environ.get("LEECH_ENABLED", "True") 

        # Will be enabled once its set
        # For vps change it to True if config loaded
        RCLONE_ENABLED = os.environ.get("RCLONE_ENABLED", "False")

        # If the user fails to select whether to use rclone or telegram to upload this will be the default.
        DEFAULT_TIMEOUT = "leech"

        # For vps set path here or you can use runtime too
        RCLONE_CONFIG = os.environ.get("RCLONE_CONFIG", "")
        
        # Name of the RCLONE drive from the config
        DEF_RCLONE_DRIVE = os.environ.get("DEF_RCLONE_DRIVE", "")

        # Max size of a playlist that is allowed (Number of videos)
        MAX_YTPLAYLIST_SIZE = int(os.environ.get("MAX_YTPLAYLIST_SIZE", "20"))
        
        # Max size of the torrent allowed
        MAX_TORRENT_SIZE = int(os.environ.get("MAX_TORRENT_SIZE", "10"))

        # Set this to your bot username if you want to add the username of your bot at the end of the commands like
        # /leech@TorToolkitBot so the value will be @TorToolkitBot
        BOT_CMD_POSTFIX = os.environ.get("BOT_CMD_POSTFIX", "")  

        # Time out for the status Delete.
        STATUS_DEL_TOUT = int(os.environ.get("STATUS_DEL_TOUT", "20"))

        # Allow the user settings to be accessed in private
        USETTINGS_IN_PRIVATE = os.environ.get("USETTINGS_IN_PRIVATE", "False") 

        # Torrent max time to collect metadata in seconds
        TOR_MAX_TOUT = int(os.environ.get("TOR_MAX_TOUT", "180"))

        # This is to stop someone from abusing the system by imposing the limit
        # [<GBs of total torrent space>, <Number of youtube videos allowed to download>, <Number of youtube playlists allowed to download>]
        USER_CAP_ENABLE = os.environ.get("USER_CAP_ENABLE", "False") 
        USER_CAP_LIMIT = [50,10,2]

        # No need to worry about these
        # CHANGE THESE AT YOUR RISK
        LOCKED_USERS = os.environ.get("LOCKED_USERS", "False") 
        RSTUFF = os.environ.get("RSTUFF", "True") 
        FORCE_DOCS_USER = os.environ.get("FORCE_DOCS_USER", "False") 
        FAST_UPLOAD = os.environ.get("FAST_UPLOAD", "True") 
        METAINFO_BOT = os.environ.get("METAINFO_BOT", "True") 
        EXPRESS_UPLOAD = os.environ.get("EXPRESS_UPLOAD", "True") 
        








# -*- coding: utf-8 -*-

from telethon.tl.types import KeyboardButtonCallback,KeyboardButton
from telethon import events
from tgtk import SessionVars
import asyncio as aio
from .getVars import get_val
from .database_handle import tkdb
from functools import partial
import time,os,configparser,logging,traceback

torlog = logging.getLogger(__name__)
#logging.getLogger("telethon").setLevel(logging.DEBUG)

TIMEOUT_SEC = 60

# this file will contian all the handlers and code for settings
# code can be more modular i think but not bothering now
# todo make the code more modular
no = "✖"
yes = "✔"
# Central object is not used its Acknowledged 
tordb = tkdb()
header =  '<b>tgtk - a telegram leech bot.</b>\n<u>administrator menu</u>'
async def handle_setting_callback(e):
    db = tordb
    session_id,_ = db.get_variable("SETTING_AUTH_CODE")
    

    data = e.data.decode()
    cmd = data.split(" ")
    val = ""
    
    if cmd[-1] != session_id:
        print("Session id",session_id," - - ",cmd[-1])
        await e.answer("this Setting menu is expired.",alert=True)
        await e.delete()
        return
    if cmd[1] == "fdocs":
        await e.answer("")
        if cmd[2] == "true":
            val = True
        else:
            val = False
        
        db.set_variable("FORCE_DOCUMENTS",val)
        SessionVars.update_var("FORCE_DOCUMENTS",val)
        await handle_settings(await e.get_message(),True,f"<b><u>changed the value to {val} of force documents.</b></u>",session_id=session_id)
    
    elif cmd[1] == "compstr":
        await e.answer("type the new value for Complete Progress String. note that only one character is expected.",alert=True)

        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e)

        await general_input_manager(e,mmes,"COMPLETED_STR","str",val[0],db,None)
        
    
    elif cmd[1] == "remstr":
        # what will a general manager require
        # anser message, type handler, value 
        await e.answer("type the new value for Remaining Progress String. note that only one character is expected.",alert=True)

        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e)
        
        await general_input_manager(e,mmes,"REMAINING_STR","str",val[0],db,None)
    
    elif cmd[1] == "tguplimit":
        # what will a general manager require
        # anser message, type handler, value 
        await e.answer("type the new value for TELEGRAM UPLOAD LIMIT. note that an integer is expected.",alert=True)

        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e)
        
        await general_input_manager(e,mmes,"TG_UP_LIMIT","int",val,db,None)

    elif cmd[1] == "maxtorsize":
        # what will a general manager require
        # answer message, type handler, value 
        await e.answer("type the new value for MAX TORRENT SIZE. note that integer an is expected.",alert=True)

        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e)
        
        await general_input_manager(e,mmes,"MAX_TORRENT_SIZE","int",val,db,None)
    
    elif cmd[1] == "maxytplsize":
        # what will a general manager require
        # answer message, type handler, value 
        await e.answer("type the new value for MAX PLAYLIST SIZE. note that an integer is expected.",alert=True)

        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e)
        
        await general_input_manager(e,mmes,"MAX_YTPLAYLIST_SIZE","int",val,db,None)

    elif cmd[1] == "rclonemenu":
        # this is menu
        mmes = await e.get_message()
        await handle_settings(mmes,True,"\nrclone config menu. TD= Team Drive, ND= Normal Drive",submenu="rclonemenu",session_id=session_id)
    elif cmd[1] == "mainmenu":
        # this is menu
        mmes = await e.get_message()
        await handle_settings(mmes,True,session_id=session_id)
    elif cmd[1] == "rcloneconfig":
        await e.answer("send the rclone config file which you have generated.",alert=True)
        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e,True)
        
        await general_input_manager(e,mmes,"RCLONE_CONFIG","str",val,db,"rclonemenu")
    elif cmd[1] == "change_drive":
        await e.answer(f"changed default drive to {cmd[2]}.",alert=True)
        db.set_variable("DEF_RCLONE_DRIVE",cmd[2])
        SessionVars.update_var("DEF_RCLONE_DRIVE",cmd[2])

        await handle_settings(await e.get_message(),True,f"<b><u>changed the default drive to {cmd[2]}</b></u>","rclonemenu",session_id=session_id)
    elif cmd[1] == "usrlock":
        if cmd[2] == "true":
            val = True
            try:
                # JIC is user does manual stuff
                await e.client.edit_permissions(e.chat_id,send_messages=False)
            except:
                await e.answer("an error occured, try again if doesn't work, report this issue.")
        else:
            try:
                await e.client.edit_permissions(e.chat_id,send_messages=True)
            except:
                await e.answer("an error occured, try again if doesn't work, report this issue.")
            val = False
        
        db.set_variable("LOCKED_USERS",val)
        SessionVars.update_var("LOCKED_USERS",val)
        await handle_settings(await e.get_message(),True,f"<b><u>changed the value to {val} of user locked.</b></u>",session_id=session_id)
        
    elif cmd[1] == "ctrlacts":
        # this is menu
        mmes = await e.get_message()
        await handle_settings(mmes,True,"\ncontrol actions menu.",submenu="ctrlacts",session_id=session_id)
    
    elif cmd[1] == "rcloneenable":
        await e.answer("note that this parameter will only work if rclone config is loaded.")
        if cmd[2] == "true":
            val = True
        else:
            val = False
        db.set_variable("RCLONE_ENABLED",val)
        SessionVars.update_var("RCLONE_ENABLED",val)
        mmes = await e.get_message()
        await handle_settings(mmes,True,f"<b><u>changed the value to {val} of rclone enabled.</b></u>","ctrlacts",session_id=session_id)
    
    elif cmd[1] == "leechenable":
        await e.answer("")
        if cmd[2] == "true":
            val = True
        else:
            val = False
        
        db.set_variable("LEECH_ENABLED",val)
        SessionVars.update_var("LEECH_ENABLED",val)
        mmes = await e.get_message()
        await handle_settings(mmes,True,f"<b><u>changed the value to {val} of leech enabled.</b></u>","ctrlacts",session_id=session_id)
    
    elif cmd[1] == "editsleepsec":
        await e.answer("type the new value for EDIT_SLEEP_SECS. note that an integer is expected.",alert=True)

        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e)
        
        await general_input_manager(e,mmes,"EDIT_SLEEP_SECS","int",val,db,None)
    elif cmd[1] == "fastupload":
        await e.answer("")

        if cmd[2] == "true":
            val = True
        else:
            val = False
        
        db.set_variable("FAST_UPLOAD",val)
        SessionVars.update_var("FAST_UPLOAD",val)
        mmes = await e.get_message()
        await handle_settings(mmes,True,f"<b><u>changed the value to {val} of fast upload enabled.</b></u>","ctrlacts",session_id=session_id)
    elif cmd[1] == "expressupload":
        await e.answer("")

        if cmd[2] == "true":
            val = True
        else:
            val = False
        
        db.set_variable("EXPRESS_UPLOAD",val)
        SessionVars.update_var("EXPRESS_UPLOAD",val)
        mmes = await e.get_message()
        await handle_settings(mmes,True,f"<b><u>changed the value to {val} of express upload enabled.</b></u>","ctrlacts",session_id=session_id)
    elif cmd[1] == "metainfo":
        await e.reply("Add @metainforobot to your group to get the metadata easily.")
    elif cmd[1] == "selfdest":
        await e.answer("closed.")
        await e.delete()
        


async def handle_settings(e,edit=False,msg="",submenu=None,session_id=None):
    # this function creates the menu
    # and now submenus too
    if session_id is None:
        session_id = time.time()
        db = tordb
        db.set_variable("SETTING_AUTH_CODE",str(session_id))
        SessionVars.update_var("SETTING_AUTH_CODE",str(session_id))
        
    
    menu = [
        #[KeyboardButtonCallback(yes+" Allow TG Files Leech123456789-","settings data".encode("UTF-8"))], # for ref
    ]
    
    if submenu is None:
        await get_bool_variable("LOCKED_USERS","Lock the Group",menu,"usrlock",session_id)
        await get_bool_variable("FORCE_DOCUMENTS","FORCE_DOCUMENTS",menu,"fdocs",session_id)
        await get_bool_variable("METAINFO_BOT","mediainforobot - get metadata of files in this group.",menu,"metainfo",session_id)
        await get_string_variable("COMPLETED_STR",menu,"compstr",session_id)
        await get_string_variable("REMAINING_STR",menu,"remstr",session_id)
        await get_int_variable("TG_UP_LIMIT",menu,"tguplimit",session_id)
        await get_int_variable("MAX_TORRENT_SIZE",menu,"maxtorsize",session_id)
        await get_int_variable("MAX_YTPLAYLIST_SIZE",menu,"maxytplsize",session_id)
        await get_int_variable("EDIT_SLEEP_SECS",menu,"editsleepsec",session_id)
        #await get_string_variable("RCLONE_CONFIG",menu,"rcloneconfig",session_id)
        await get_sub_menu("open rclone menu","rclonemenu",session_id,menu)
        await get_sub_menu("control actions","ctrlacts",session_id,menu)
        menu.append(
            [KeyboardButtonCallback("close menu",f"settings selfdest {session_id}".encode("UTF-8"))]
        )


        if edit:
            rmess = await e.edit(header+"\nit is recommended to lock the group before setting vars.\n"+msg,parse_mode="html",buttons=menu,link_preview=False, file="tk.jpg")
        else:
            rmess = await e.reply(header+"\nit is recommended to lock the group before setting vars.\n",parse_mode="html",buttons=menu,link_preview=False, file="tk.jpg")
    elif submenu == "rclonemenu":
        rcval = await get_string_variable("RCLONE_CONFIG",menu,"rcloneconfig",session_id)
        if rcval != "None":
            # create a all drives menu
            if rcval == "Custom file is loaded.":
                db = tordb
                _, fdata = db.get_variable("RCLONE_CONFIG")
                
                path = os.path.join(os.getcwd(),"rclone.conf")

                # find alternative to this
                with open(path,"wb") as fi:
                    fi.write(fdata)
                
                conf = configparser.ConfigParser()
                conf.read(path)
                #menu.append([KeyboardButton("Choose a default drive from below")])
                def_drive = get_val("DEF_RCLONE_DRIVE")

                for j in conf.sections():
                    prev=""
                    if j == def_drive:
                        prev = yes

                    if "team_drive" in list(conf[j]):
                        menu.append(
                            [KeyboardButtonCallback(f"{prev}{j} - TD",f"settings change_drive {j} {session_id}")]
                        )
                    else:
                        menu.append(
                            [KeyboardButtonCallback(f"{prev}{j} - ND",f"settings change_drive {j} {session_id}")]
                        )

        await get_sub_menu("go back ⬅️","mainmenu",session_id,menu)
        menu.append(
            [KeyboardButtonCallback("close menu",f"settings selfdest {session_id}".encode("UTF-8"))]
        )
        if edit:
            rmess = await e.edit(header+"\nit's recommended to lock the group before setting vars.\n"+msg,parse_mode="html",buttons=menu,link_preview=False)

    elif submenu == "ctrlacts":
        await get_bool_variable("RCLONE_ENABLED","enable rclone.",menu,"rcloneenable",session_id)
        await get_bool_variable("LEECH_ENABLED","enable leech.",menu,"leechenable",session_id)
        await get_bool_variable("FAST_UPLOAD","enable fast upload. (turn off if you get errors)",menu,"fastupload",session_id)
        await get_bool_variable("EXPRESS_UPLOAD","enable express upload. (turn off if you get errors.)",menu,"expressupload",session_id)
        await get_bool_variable("FORCE_DOCS_USER","force_docs_user - not implemented.",menu,"forcedocsuser",session_id)


        await get_sub_menu("go back ⬅️","mainmenu",session_id,menu)
        menu.append(
            [KeyboardButtonCallback("close menu",f"settings selfdest {session_id}".encode("UTF-8"))]
        )
        if edit:
            rmess = await e.edit(header+"\nit is recommended to lock the group before setting vars.\n"+msg,parse_mode="html",buttons=menu,link_preview=False)

# an attempt to manager all the input
async def general_input_manager(e,mmes,var_name,datatype,value,db,sub_menu):
    if value is not None and not "ignore" in value:
        await confirm_buttons(mmes,value)
        conf = await get_confirm(e)
        if conf is not None:
            if conf:
                try:
                    if datatype == "int":
                        value = int(value)
                    if datatype == "str":
                        value = str(value)
                    if datatype == "bool":
                        if value.lower() == "true":
                            value = True
                        elif value.lower() == "false":
                            value = False
                        else:
                            raise ValueError("Invalid value from bool")
                    
                    if var_name == "RCLONE_CONFIG":
                        #adjust the special case
                        try:
                            conf = configparser.ConfigParser()
                            conf.read(value)
                            for i in conf.sections():
                                db.set_variable("DEF_RCLONE_DRIVE",str(i))
                                SessionVars.update_var("DEF_RCLONE_DRIVE",str(i))
                                break
                                
                            with open(value,"rb") as fi:
                                data = fi.read()
                                db.set_variable("RCLONE_CONFIG",0,True,data)
                            os.remove(value)
                            db.set_variable("LEECH_ENABLED",True)
                            SessionVars.update_var("LEECH_ENABLED",True)
                        except Exception:
                            torlog.error(traceback.format_exc())
                            await handle_settings(mmes,True,f"<b><u>the configuration file is invalid, check logs.</b></u>",sub_menu)
                            return
                    else:
                        db.set_variable(var_name,value)
                        SessionVars.update_var(var_name,value)
                    
                    await handle_settings(mmes,True,f"<b><u>received {var_name} value '{value}' with confirm.</b></u>",sub_menu)
                except ValueError:
                    await handle_settings(mmes,True,f"<b><u>value [{value}] not valid try again and enter {datatype}.</b></u>",sub_menu)    
            else:
                await handle_settings(mmes,True,f"<b><u>confirm differed by user.</b></u>",sub_menu)
        else:
            await handle_settings(mmes,True,f"<b><u>confirm timed out [waited 60s for input].</b></u>",sub_menu)
    else:
        await handle_settings(mmes,True,f"<b><u>entry timed out [waited 60s for input].</b></u>",sub_menu)


async def get_value(e,file=False):
    # todo replace with conver. - or maybe not Fix Dont switch to conversion
    # this function gets the new value to be set from the user in current context
    lis = [False,None]

    #func tools works as expected ;);)    
    cbak = partial(val_input_callback,o_sender=e.sender_id,lis=lis,file=file)
    
    e.client.add_event_handler(
        #lambda e: test_callback(e,lis),
        cbak,
        events.NewMessage()
    )

    start = time.time()

    while not lis[0]:
        if (time.time() - start) >= TIMEOUT_SEC:
            break

        await aio.sleep(1)
    
    val = lis[1]
    
    e.client.remove_event_handler(cbak)

    return val

async def get_confirm(e):
    # abstract for getting the confirm in a context

    lis = [False,None]
    cbak = partial(get_confirm_callback,o_sender=e.sender_id,lis=lis)
    
    e.client.add_event_handler(
        #lambda e: test_callback(e,lis),
        cbak,
        events.CallbackQuery(pattern="confirmsetting")
    )

    start = time.time()

    while not lis[0]:
        if (time.time() - start) >= TIMEOUT_SEC:
            break
        await aio.sleep(1)

    val = lis[1]
    
    e.client.remove_event_handler(cbak)

    return val

async def val_input_callback(e,o_sender,lis,file):
    # get the input value
    if o_sender != e.sender_id:
        return
    if not file:
        lis[0] = True
        lis[1] = e.text
        await e.delete()
    else:
        if e.document is not None:
            path = await e.download_media()
            lis[0]  = True
            lis[1] = path 
            await e.delete()
        else:
            if "ignore" in e.text:
                lis[0]  = True
                lis[1] = "ignore"
                await e.delete()
            else:
                await e.delete()
        
    raise events.StopPropagation

async def get_confirm_callback(e,o_sender,lis):
    # handle the confirm callback

    if o_sender != e.sender_id:
        return
    lis[0] = True
    
    data = e.data.decode().split(" ")
    if data[1] == "true":
        lis[1] = True
    else:
        lis[1] = False

async def confirm_buttons(e,val):
    # add the confirm buttons at the bottom of the message
    await e.edit(f"confirm the input: <u>{val}</u>",buttons=[KeyboardButtonCallback("yes","confirmsetting true"),KeyboardButtonCallback("no","confirmsetting false")],parse_mode="html")

async def get_bool_variable(var_name,msg,menu,callback_name,session_id):
    # handle the vars having bool values
     
    val = get_val(var_name)
    
    if val:
        #setting the value in callback so calls will be reduced ;)
        menu.append(
            [KeyboardButtonCallback(yes+msg,f"settings {callback_name} false {session_id}".encode("UTF-8"))]
        ) 
    else:
        menu.append(
            [KeyboardButtonCallback(no+msg,f"settings {callback_name} true {session_id}".encode("UTF-8"))]
        ) 

async def get_sub_menu(msg,sub_name,session_id,menu):
    menu.append(
        [KeyboardButtonCallback(msg,f"settings {sub_name} {session_id}".encode("UTF-8"))]
    )

async def get_string_variable(var_name,menu,callback_name,session_id):
    # handle the vars having string value
    # condition for rclone config

    val = get_val(var_name)
    if var_name == "RCLONE_CONFIG":
        db = tordb
        _, val1 = db.get_variable(var_name)
        if val1 is not None:
            val = "custom file is loaded."
        
    msg = var_name + " " + str(val)
    menu.append(
        [KeyboardButtonCallback(msg,f"settings {callback_name} {session_id}".encode("UTF-8"))]
    )

    # Just in case
    return val

async def get_int_variable(var_name,menu,callback_name,session_id):
    # handle the vars having string value

    val = get_val(var_name)
    msg = var_name + " " + str(val)
    menu.append(
        [KeyboardButtonCallback(msg,f"settings {callback_name} {session_id}".encode("UTF-8"))]
    ) 

# todo handle the list value 



# -*- coding: utf-8 -*-


from urllib.parse import urlparse,parse_qs
import logging
from torrentool.api import Torrent

#logging.basicConfig(level=logging.DEBUG)

def get_hash_magnet(mgt):
    if mgt.startswith('magnet:'):
        _, _, _, _, query, _ = urlparse(mgt)

    qs = parse_qs(query)
    v = qs.get('xt', None)
    
    if v == None or v == []:
        logging.error('Invalid magnet URI: no "xt" query parameter.')
        return False
        
    v = v[0]
    if not v.startswith('urn:btih:'):
        logging.error('Invalid magnet URI: "xt" value not valid for BitTorrent.')
        return False

    mgt = v[len('urn:btih:'):]
    return mgt.lower()

def get_hash_file(path):
    tr = Torrent.from_file(path)
    mgt = tr.magnet_link
    return get_hash_magnet(mgt)


import re
import aiohttp
import math
import logging
from bs4 import BeautifulSoup

torlog = logging.getLogger(__name__)

async def generate_directs(url):
    #Mediafire
    if 'mediafire.com' in url:
        try:
            link = re.findall(r'\bhttps?://.*mediafire\.com\S+', url)[0]
            async with aiohttp.ClientSession() as ttksess:
                resp = await ttksess.get(link)
                restext = await resp.text()
            
            page = BeautifulSoup(restext, 'lxml')
            info = page.find('a', {'aria-label': 'Download file'})
            ourl = info.get('href')
            return ourl
        except:
            return "**ERROR No mediafire link found.**"
        
        
    #Zippyshare
    elif  'zippyshare.com' in url:
        
        #Zippyshare fix directlink generator for ttk by @yourtulloh based on:
        #Zippyshare up-to-date plugin from https://github.com/UsergeTeam/Userge-Plugins/blob/master/plugins/zippyshare.py
        #Thanks to all contributors @aryanvikash, rking32, @BianSepang
        tulloh = r'https://www(\d{1,3}).zippyshare.com/v/(\w{8})/file.html'
        regex_result = (
            r'var a = (\d{6});\s+var b = (\d{6});\s+document\.getElementById'
            r'\(\'dlbutton\'\).omg = "f";\s+if \(document.getElementById\(\''
            r'dlbutton\'\).omg != \'f\'\) {\s+a = Math.ceil\(a/3\);\s+} else'
            r' {\s+a = Math.floor\(a/3\);\s+}\s+document.getElementById\(\'d'
            r'lbutton\'\).href = "/d/[a-zA-Z\d]{8}/\"\+\(a \+ \d{6}%b\)\+"/('
            r'[\w%-.]+)";'
        )
        
        match = re.match(tulloh, url)
        if not match:
            return "**ERROR Invalid Zippyshare URL.**"
        
        server, id_ = match.group(1), match.group(2)
        async with aiohttp.ClientSession() as ttksess:
            resp = await ttksess.get(url)
            restext = await resp.text()
            
        match = re.search(regex_result, restext, re.DOTALL)
        if not match:
            return "**ERROR Invalid response, try again!**"
        
        val_1 = int(match.group(1))
        val_2 = math.floor(val_1 / 3)
        val_3 = int(match.group(2))
        val = val_1 + val_2 % val_3
        name = match.group(3)
        ourl = "https://www{}.zippyshare.com/d/{}/{}/{}".format(server, id_, val, name)
        return ourl
        
    elif 'yadi.sk' in url or 'disk.yandex.com' in url:
        try:
            link = re.findall(r'\b(https?://.*(yadi|disk)\.(sk|yandex)*(|com)\S+)', url)[0][0]
            print(link)
        except:
            return "**ERROR No yadisk link found.**"

        api = 'https://cloud-api.yandex.net/v1/disk/public/resources/download?public_key={}'
        try:
            async with aiohttp.ClientSession() as ttksess:
                resp = await ttksess.get(api.format(link))
                restext = await resp.json()
                ourl = restext['href']
                return ourl
        except:
            torlog.exception("Ayee jooo")
            return "**ERROR 404 File Not Found** or \n**Download limit reached.**"


# -*- coding: utf-8 -*-


import re,os,shutil,time, aiohttp
from telethon.tl import types
import logging, shutil
import math, json
import urllib.parse
import asyncio as aio
from bs4 import BeautifulSoup
from . import QBittorrentWrap
from . import ariatools
from .tele_upload import upload_handel
from .rclone_upload import rclone_driver
from .zip7_utils import add_to_zip, extract_archive
from ..core.getVars import get_val
from ..core.status.status import ARTask
from ..core.status.upload import TGUploadTask
from ..functions.Human_Format import human_readable_bytes
from .dl_generator import generate_directs

#logging.basicConfig(level=logging.DEBUG)
logging.getLogger("telethon").setLevel(logging.WARNING)
torlog = logging.getLogger(__name__)

#this files main task is to keep the ability to switch to a new engine if needed ;)

#TODO implement multiple magnets from same message if needed
#this function is to ensure that only one magnet is passed at a time
def get_magnets(text):
    matches = [ i for i in re.finditer("magnet:",text)]
    magnets = list()

    for i in range(len(matches)):
        if i == len(matches)-1:
            magnets.append(text[matches[i].span()[0]:])
        elif i == 0:
            magnets.append(text[:matches[i+1].span()[0]])
        else:
            magnets.append(text[matches[i].span()[0]:matches[i+1].span()[0]])

    for i in range(len(magnets)):
        magnets[i] = magnets[i].strip()

    return magnets[0]


def get_entities(msg):
    urls = list()

    for i in msg.entities:
        if isinstance(i,types.MessageEntityUrl):
            o,l = i.offset,i.length
            urls.append(msg.text[o:o+l])
        elif isinstance(i,types.MessageEntityTextUrl):
            urls.append(i.url)

    if len(urls) > 0:
        return urls[0]
    else:
        return None

async def check_link(msg,rclone=False,is_zip=False, extract=False, prev_msg=None):
    # here moslty rmess = Reply message which the bot uses to update
    # omess = original message from the sender user 
    omess = msg
    msg = await msg.get_reply_message()

    if extract:
        mess = f"you chose to extract the archive <a href='tg://user?id={omess.sender_id}'>enter password if any.</a>\n use <code>/setpass {omess.id} password-here</code>"
        omess.client.dl_passwords[omess.id] = [str(omess.sender_id), None]
        await omess.reply(mess, parse_mode="html")

    if msg is None:
        urls = None

    elif msg.document is not None:
        name = None
        for i in msg.document.attributes:
            if isinstance(i,types.DocumentAttributeFilename):
                name = i.file_name
        
        if name is None:
            await omess.reply("this isn't a torrent file to leech from. send <code>.torrent</code> file",parse_mode="html")
        elif name.lower().endswith(".torrent"):
            rmess = await omess.reply("downloading the torrent file.")

            #not worring about the download location now
            # TODO do something to de register the torrents
            path = await msg.download_media()
            torrent_return =  await QBittorrentWrap.register_torrent(path,rmess,omess,file=True)
            
            dl_path, dl_task = None, None
            if not isinstance(torrent_return, bool) and torrent_return is not None:
                dl_path = torrent_return[0]
                dl_task = torrent_return[1]

                if extract:
                    newpath = await handle_ext_zip(dl_path, rmess, omess)
                    if not newpath is False:
                        dl_path = newpath
                else:
                    newpath = await handle_zips(dl_path, is_zip, rmess, not rclone)
                    if newpath is False:
                        pass
                    else:
                        dl_path = newpath
                
                if not rclone:
                    ul_task = TGUploadTask(dl_task)
                    await ul_task.dl_files()
                    try:
                        rdict = await upload_handel(dl_path,rmess,omess.from_id,dict(),user_msg=omess,task=ul_task)
                    except:
                        rdict = dict()
                        torlog.exception("exception in torrent file")
                    
                    await ul_task.set_inactive()
                    await print_files(omess,rdict,dl_task.hash, path = dl_path)
                    torlog.info("here are the files uploaded {}".format(rdict))
                    await QBittorrentWrap.delete_this(dl_task.hash)
                else:
                    res = await rclone_driver(dl_path,rmess,omess, dl_task)
                    if res is None:
                        await msg.reply("<b>upload to drive failed. check logs for more info./b>",parse_mode="html")
                    await QBittorrentWrap.delete_this(dl_task.hash)
            
            else:
                await errored_message(omess, rmess)

            await clear_stuff(path)
            await clear_stuff(dl_path)
            return dl_path
        else:
            await omess.reply("this is not a torrent file to leech from. send <code>.torrent</code> file.",parse_mode="html")

    elif msg.raw_text is not None:
        if msg.raw_text.lower().startswith("magnet:"):
            rmess = await omess.reply("scanning....")
            
            mgt = get_magnets(msg.raw_text.strip())
            torrent_return = await QBittorrentWrap.register_torrent(mgt,rmess,omess,True)
            
            dl_path, dl_task = None, None
            if not isinstance(torrent_return,bool) and torrent_return is not None:
                dl_path = torrent_return[0]
                dl_task = torrent_return[1]
                
                if extract:
                    newpath = await handle_ext_zip(dl_path, rmess, omess)
                    if not newpath is False:
                        dl_path = newpath
                else:
                    newpath = await handle_zips(dl_path, is_zip, rmess, not rclone)
                    if newpath is False:
                        pass
                    else:
                        dl_path = newpath

                if not rclone:
                    # TODO add exception update for tg upload everywhere
                    
                    ul_task = TGUploadTask(dl_task)
                    await ul_task.dl_files()
                    
                    try:
                        rdict = await upload_handel(dl_path,rmess,omess.from_id,dict(),user_msg=omess,task=ul_task)
                    except:
                        rdict = dict()
                        torlog.exception("exception in magnet")
                    
                    await ul_task.set_inactive()
                    await print_files(omess,rdict,dl_task.hash, path = dl_path)
                    
                    torlog.info("here are the files to be uploaded {}".format(rdict))
                    await QBittorrentWrap.delete_this(dl_task.hash)

                else:
                    res = await rclone_driver(dl_path,rmess,omess, dl_task)
                    if res is None:
                        await msg.reply("<b>upload to drive failed. check logs for more info./b>",parse_mode="html")
                    await QBittorrentWrap.delete_this(dl_task.hash)
            else:
                await errored_message(omess, rmess)
            
            await clear_stuff(dl_path)

        elif msg.raw_text.lower().endswith(".torrent"):
            rmess = await omess.reply("downloading the torrent file.")

            # TODO do something to de register the torrents - done
            path = ""
            async with aiohttp.ClientSession() as sess:
                async with sess.get(msg.raw_text) as resp:
                    if resp.status == 200:
                        path = str(time.time()).replace(".","")+".torrent"
                        with open(path, "wb") as fi:
                            fi.write(await resp.read())
                    else:
                        await rmess.edit("error got http response code:- "+str(resp.status))
                        return    

            torrent_return =  await QBittorrentWrap.register_torrent(path,rmess,omess,file=True)
            dl_path, dl_task = None, None
            if not isinstance(torrent_return,bool) and torrent_return is not None:
                dl_path = torrent_return[0]
                dl_task = torrent_return[1]

                if extract:
                    newpath = await handle_ext_zip(dl_path, rmess, omess)
                    if not newpath is False:
                        dl_path = newpath
                else:
                    newpath = await handle_zips(dl_path, is_zip, rmess, not rclone)
                    if newpath is False:
                        pass
                    else:
                        dl_path = newpath
                        
                if not rclone:
                    ul_task = TGUploadTask(dl_task)
                    await ul_task.dl_files()

                    try:
                        rdict = await upload_handel(dl_path,rmess,omess.from_id,dict(),user_msg=omess,task=ul_task)
                    except:
                        rdict = dict()
                        torlog.exception("exception in torrent link")
                    
                    await ul_task.set_inactive()
                    await print_files(omess,rdict,dl_task.hash, path = dl_path)
                    
                    torlog.info("here are the files uploaded {}".format(rdict))
                    await QBittorrentWrap.delete_this(dl_task.hash)
                else:
                    res = await rclone_driver(dl_path,rmess,omess, dl_task)
                    if res is None:
                        await msg.reply("<b>upload to drive failed. check logs for more info.</b>",parse_mode="html")
                    await QBittorrentWrap.delete_this(dl_task.hash)
            else:
                await errored_message(omess, rmess)

            await clear_stuff(path)
            await clear_stuff(dl_path)
            return dl_path

        else:
            urls = msg.raw_text
            url = msg.raw_text
            torlog.info("aria2 is downloading:\n{}".format(urls))
            rmsg = await omess.reply("**processing the link...**")
            await aio.sleep(1)
            
            url = await generate_directs(urls)
            if url is not None:
                if "**ERROR" in url:
                    await omess.reply(url)
                    return

            re_name = None
            try:
                if " "  in omess.raw_text:
                    cmd = omess.raw_text.split(" ", 1)[-1]
                    if len(cmd) > 0:
                        re_name = cmd
                    else:
                        torlog.info(f"this is not a valid name for renaming: {omess.raw_text}")
            except:
                torlog.exception("invaild rename detected.")
            
            # weird stuff had to refect message
            path = None
            rmsg = await omess.client.get_messages(ids=rmsg.id, entity=rmsg.chat_id)

            if url is None:
                stat, dl_task = await ariatools.aria_dl(msg.raw_text,"",rmsg,omess)
            else:
                stat, dl_task = await ariatools.aria_dl(url,"",rmsg,omess)
            if isinstance(dl_task,ARTask) and stat:
                path = await dl_task.get_path()
                if re_name:
                    try:
                        rename_path = os.path.join(os.path.dirname(path), re_name)
                        os.rename(path, rename_path)
                        path = rename_path
                    except:
                        torlog.warning("invaild name for file.")

                if extract:
                    newpath = await handle_ext_zip(path, rmsg, omess)
                    if not newpath is False:
                        path = newpath
                else:
                    newpath = await handle_zips(path, is_zip, rmsg, not rclone)
                    if newpath is False:
                        pass
                    else:
                        path = newpath
                
                if not rclone:
                    ul_task = TGUploadTask(dl_task)
                    await ul_task.dl_files()
                    
                    try:
                        rdict = await upload_handel(path,rmsg,omess.from_id,dict(),user_msg=omess,task=ul_task)
                    except:
                        rdict = dict()
                        torlog.exception("exception in direct links.")
                    
                    await ul_task.set_inactive()
                    await print_files(omess,rdict, path = path)
                    torlog.info("here are the files to be uploaded {}".format(rdict))
                else:
                    res = await rclone_driver(path,rmsg, omess, dl_task)
                    if res is None:
                        await msg.reply("<b>upload to drive failed. check logs for more info.</b>",parse_mode="html")
            elif stat is False:
                reason = await dl_task.get_error()
                await rmsg.edit("failed to download this file.\n"+str(reason))
                await errored_message(omess, rmsg)
            
            await clear_stuff(path)    
    return None

async def pause_all(msg):
    await QBittorrentWrap.pause_all(msg)

async def resume_all(msg):
    await QBittorrentWrap.resume_all(msg)

async def purge_all(msg):
    await QBittorrentWrap.delete_all(msg)

async def get_status(msg,all=False):
    smsg = await QBittorrentWrap.get_status(msg,all)
    if len(smsg) > 3600:
        chunks, chunk_size = len(smsg), len(smsg)//4
        msgs = [ smsg[i:i+chunk_size] for i in range(0, chunks, chunk_size) ]
        
        for j in msgs:
            await msg.reply(j,parse_mode="html")
            await aio.sleep(1)
    else:
        await msg.reply(smsg,parse_mode="html")

async def handle_zips(path, is_zip, rmess, split=True):
    # refetch rmess
    rmess = await rmess.client.get_messages(rmess.chat_id,ids=rmess.id)
    if is_zip:
        try:
            await rmess.edit(rmess.text+"\n starting to archive the contents. please wait.")
            zip_path = await add_to_zip(path, get_val("TG_UP_LIMIT"), split)
            
            if zip_path is None:
                await rmess.edit(rmess.text+"\n zip failed. fallback to normal")
                return False
            
            if os.path.isdir(path):
                shutil.rmtree(path)
            if os.path.isfile(path):
                os.remove(path)
            await rmess.edit(rmess.text+"\n zipping completed, now uploading.")
            await clear_stuff(path)
            return zip_path
        except:
            await rmess.edit(rmess.text+"\n zip failed. fallback to normal")
            return False
    else:
        return path

async def handle_ext_zip(path, rmess, omess):
    # refetch rmess
    rmess = await rmess.client.get_messages(rmess.chat_id,ids=rmess.id)
    password = rmess.client.dl_passwords.get(omess.id)
    if password is not None:
        password = password[1]
    start = time.time()
    await rmess.edit(f"{rmess.text} trying to extract the archive with password <code>{password}</code>.", parse_mode="html")
    wrong_pwd = False

    while True:
        if not wrong_pwd:
            ext_path = await extract_archive(path,password=password)
        else:
            if (time.time() - start) > 1200:
                await rmess.edit(f"{rmess.text} extract failed as no correct password was provided, uploading as it is.")
                return False

            temppass = rmess.client.dl_passwords.get(omess.id)
            if temppass is not None:
                temppass = temppass[1]
            if temppass == password:
                await aio.sleep(10)
                continue
            else:
                password = temppass
                wrong_pwd = False
                continue
        
        if isinstance(ext_path, str):
            if "Wrong Password" in ext_path:
                mess = f"<a href='tg://user?id={omess.sender_id}'>re-enter password</a>\nthe password <code>{password}</code> you provided is an incorrect password. you have {((time.time()-start)/60)-20} mins to reply or else, the zip will be uploaded as is.\n use <code>/setpass {omess.id} password-here</code>"
                await omess.reply(mess, parse_mode="html")
                wrong_pwd = True
            else:
                await clear_stuff(path)
                return ext_path
        
        elif ext_path is False:
            return False
        elif ext_path is None:
            # None is to descibe fetal but the upload will fail 
            # itself further nothing to handle here
            return False
        else:
            await clear_stuff(path)
            return ext_path
            
async def errored_message(e, reason):
    msg = f"<a href='tg://user?id={e.sender_id}'>completed.</a>\ndownload failed."
    if reason is not None:
        await reason.reply(msg, parse_mode="html")
    else:
        await e.reply(msg, parse_mode="html")

async def print_files(e,files,thash=None, path = None):
    msg = f"<a href='tg://user?id={e.sender_id}'>completed.</a>"

    if path is not None:
        size = 0
        try:
            if os.path.isdir(path):
                size = get_size_fl(path)
            else:
                size = os.path.getsize(path)
        except:
            torlog.warning("size calculation failed.")
        
        size = human_readable_bytes(size)
        msg += f"uploaded size: {str(size)}\n\n"
    
    if len(files) == 0:
        return
    
    chat_id = e.chat_id
    msg_li = []
    for i in files.keys():
        link = f'https://t.me/c/{str(chat_id)[4:]}/{files[i]}'
        if len(msg + f'<a href="{link}">{i}</a>\n') > 4000:
            msg_li.append(msg)
            msg = f'<a href="{link}">{i}</a>\n'
        else:
            msg += f'<a href="{link}">{i}</a>\n'

    for i in msg_li:
        await e.reply(i,parse_mode="html")
        await aio.sleep(1)
        
    await e.reply(msg,parse_mode="html")

    try:
        if thash is not None:
            from .store_info_hash import store_driver # pylint: disable=import-error
            await store_driver(e, files, thash) 
    except:
        pass

    if len(files) < 2:
        return

    ids = list()
    for i in files.keys():
        ids.append(files[i])
    
    msgs = await e.client.get_messages(e.chat_id,ids=ids)
    for i in msgs:
        index = None
        for j in range(0,len(msgs)):
            index = j
            if ids[j] == i.id:
                break
        nextt,prev = "",""
        chat_id = str(e.chat_id)[4:]
        buttons = []
        if index == 0:
            nextt = f'https://t.me/c/{chat_id}/{ids[index+1]}'
            buttons.append(
                types.KeyboardButtonUrl("next", nextt)
            )
            nextt = f'<a href="{nextt}">next</a>\n'
        elif index == len(msgs)-1:
            prev = f'https://t.me/c/{chat_id}/{ids[index-1]}'
            buttons.append(
                types.KeyboardButtonUrl("prev", prev)
            )
            prev = f'<a href="{prev}">prev</a>\n'
        else:
            nextt = f'https://t.me/c/{chat_id}/{ids[index+1]}'
            buttons.append(
                types.KeyboardButtonUrl("next", nextt)
            )
            nextt = f'<a href="{nextt}">next</a>\n'
            
            prev = f'https://t.me/c/{chat_id}/{ids[index-1]}'
            buttons.append(
                types.KeyboardButtonUrl("prev", prev)
            )
            prev = f'<a href="{prev}">prev</a>\n'

        try:
            #await i.edit("{} {} {}".format(prev,i.text,nextt),parse_mode="html")
            await i.edit(buttons=buttons)
        except:pass
        await aio.sleep(2)
    


async def get_transfer():
    client = await QBittorrentWrap.get_client()
    data = client.transfer_info()
    print(data)
    return data

async def clear_stuff(path):
    try:
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)
    except:pass


async def cancel_torrent(hashid, is_aria = False):
    if not is_aria:
        await QBittorrentWrap.deregister_torrent(hashid)
    else:
        await ariatools.remove_dl(hashid)

def get_size_fl(start_path = '.'):
    total_size = 0
    for dirpath, _, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size



# -*- coding: utf-8 -*-


import os,logging,time,traceback,shutil, asyncio
from ..core.getVars import get_val
from ..core import thumb_manage # i guess i will dodge this one ;) as i am importing the vids helper anyways
from . import vids_helpers,zip7_utils
from .progress_for_telethon import progress
from hachoir.parser import createParser
from hachoir.metadata import extractMetadata
from telethon.errors import VideoContentTypeInvalidError
from ..core.database_handle import tkupload
from .. import user_db
from telethon.tl.types import KeyboardButtonCallback,DocumentAttributeVideo,DocumentAttributeAudio
from telethon.utils import get_attributes
from .Ftele import upload_file
from pyrogram.types import InputMediaDocument, InputMediaVideo, InputMediaPhoto, InputMediaAudio, InlineKeyboardButton, InlineKeyboardMarkup
from .progress_for_pyrogram import progress_for_pyrogram
torlog = logging.getLogger(__name__)

#thanks @SpEcHiDe for this concept of recursion
async def upload_handel(path,message,from_uid,files_dict,job_id=0,force_edit=False,updb=None,from_in=False,thumb_path=None, user_msg=None, task=None):
    # creting here so connections are kept low
    if updb is None:
        # Central object is not used its Acknowledged 
        updb = tkupload()

    #logging.info("Uploading Now:- {}".format(path))

    if os.path.isdir(path):
        logging.info("uploading the directory: {}".format(path))

        directory_contents = os.listdir(path)
        directory_contents.sort()
        try:
            # maybe way to refresh?!
            message = await message.client.get_messages(message.chat_id,ids=[message.id])
            message = message[0]
        except:pass

        try:
            message = await message.edit("{}\nfound **{}** files for this download.".format(message.text,len(directory_contents)))
        except:
        	torlog.warning("too much folders, will stop the editing of this message.")
        
        if not from_in:
            updb.register_upload(message.chat_id,message.id)
            if user_msg is None:
                sup_mes = await message.get_reply_message()
            else:
                sup_mes = user_msg
            
            if task is not None:
                await task.set_message(message)
                await task.set_original_message(sup_mes)
            
            data = "upcancel {} {} {}".format(message.chat_id,message.id,sup_mes.sender_id)
            buts = [KeyboardButtonCallback("cancel upload.",data.encode("UTF-8"))]
            message = await message.edit(buttons=buts)


        for file in directory_contents:
            if updb.get_cancel_status(message.chat_id,message.id):
                continue

            await upload_handel(
                os.path.join(path,file),
                message,
                from_uid,
                files_dict,
                job_id,
                force_edit,
                updb,
                from_in=True,
                thumb_path=thumb_path,
                user_msg=user_msg,
                task=task
            )
        
        if not from_in:
            if updb.get_cancel_status(message.chat_id,message.id):
                task.cancel = True
                await task.set_inactive()
                await message.edit("`{}` - canceled by user.".format(message.text),buttons=None)
            else:
                await message.edit(buttons=None)
            updb.deregister_upload(message.chat_id,message.id)

    else:
        logging.info("uploading the file: {}".format(path))
        if os.path.getsize(path) > get_val("TG_UP_LIMIT"):
            # the splitted file will be considered as a single upload ;)
            
            
            metadata = extractMetadata(createParser(path))
            
            if metadata is not None:
                # handle none for unknown
                metadata = metadata.exportDictionary()
                try:
                    mime = metadata.get("Common").get("MIME type")
                except:
                    mime = metadata.get("Metadata").get("MIME type")

                ftype = mime.split("/")[0]
                ftype = ftype.lower().strip()
            else:
                ftype = "unknown"
            
            if ftype == "video":    
                todel = await message.reply("**file larger then threshold, splitting now. **\n**processing.....** ```using algo ffmpeg split```") 
                split_dir = await vids_helpers.split_file(path,get_val("TG_UP_LIMIT"))
                await todel.delete()
            else:
                todel = await message.reply("**file larger then threshold, splitting now. **\n**processing.....** ```using algo ffmpeg split```")  
                split_dir = await zip7_utils.split_in_zip(path,get_val("TG_UP_LIMIT"))
                await todel.delete()
            
            if task is not None:
                await task.add_a_dir(split_dir)
            
            dircon = os.listdir(split_dir)
            dircon.sort()

            if not from_in:
                updb.register_upload(message.chat_id,message.id)
                if user_msg is None:
                    sup_mes = await message.get_reply_message()
                else:
                    sup_mes = user_msg

                if task is not None:
                    await task.set_message(message)
                    await task.set_original_message(sup_mes)

                data = "upcancel {} {} {}".format(message.chat_id,message.id,sup_mes.sender_id)
                buts = [KeyboardButtonCallback("cancel upload.",data.encode("UTF-8"))]
                await message.edit(buttons=buts)

            for file in dircon:
                if updb.get_cancel_status(message.chat_id,message.id):
                    continue
            
                await upload_handel(
                    os.path.join(split_dir,file),
                    message,
                    from_uid,
                    files_dict,
                    job_id,
                    force_edit,
                    updb=updb,
                    from_in=True,
                    thumb_path=thumb_path,
                    user_msg=user_msg,
                    task=task
                )
            
            try:
                shutil.rmtree(split_dir)
                os.remove(path)
            except:pass
            
            if not from_in:
                if updb.get_cancel_status(message.chat_id,message.id):
                    task.cancel = True
                    await task.set_inactive()
                    await message.edit("`{}` - canceled By user.".format(message.text),buttons=None)
                else:
                    await message.edit(buttons=None)
                updb.deregister_upload(message.chat_id,message.id)
            # spliting file logic blah blah
        else:
            if not from_in:
                updb.register_upload(message.chat_id,message.id)
                if user_msg is None:
                    sup_mes = await message.get_reply_message()
                else:
                    sup_mes = user_msg
                
                if task is not None:
                    await task.set_message(message)
                    await task.set_original_message(sup_mes)

                if task is not None:
                    await task.set_message(message)
                    await task.set_original_message(sup_mes)

                data = "upcancel {} {} {}".format(message.chat_id,message.id,sup_mes.sender_id)
                buts = [KeyboardButtonCallback("cancel upload.",data.encode("UTF-8"))]
                await message.edit(buttons=buts)
            #print(updb)
            if black_list_exts(path):
                if task is not None:
                    await task.uploaded_file(os.path.basename(path))
                sentmsg = None
            else:
                sentmsg = await upload_a_file(
                    path,
                    message,
                    force_edit,
                    updb,
                    thumb_path,
                    user_msg=user_msg
                )

            if not from_in:
                if updb.get_cancel_status(message.chat_id,message.id):
                    task.cancel = True
                    await task.set_inactive()
                    await message.edit("`{}` - canceled By user.".format(message.text),buttons=None)
                else:
                    await message.edit(buttons=None)
                updb.deregister_upload(message.chat_id,message.id)

            if sentmsg is not None:
                if task is not None:
                    await task.uploaded_file(os.path.basename(path))
                files_dict[os.path.basename(path)] = sentmsg.id

    return files_dict



async def upload_a_file(path,message,force_edit,database=None,thumb_path=None,user_msg=None):
    if get_val("EXPRESS_UPLOAD"):
        return await upload_single_file(path, message, force_edit, database, thumb_path, user_msg)
    queue = message.client.queue
    if database is not None:
        if database.get_cancel_status(message.chat_id,message.id):
            # add os remove here
            return None
    if not os.path.exists(path):
        return None
        
    if user_msg is None:
        user_msg = await message.get_reply_message()
    
    #todo improve this uploading ✔️
    file_name = os.path.basename(path)
    caption_str = ""
    caption_str += "<code>"
    caption_str += file_name
    caption_str += "</code>"
    metadata = extractMetadata(createParser(path))
    ometa = metadata
    
    if metadata is not None:
        # handle none for unknown
        metadata = metadata.exportDictionary()
        try:
            mime = metadata.get("Common").get("MIME type")
        except:
            mime = metadata.get("Metadata").get("MIME type")

        ftype = mime.split("/")[0]
        ftype = ftype.lower().strip()
    else:
        ftype = "unknown"
    #print(metadata)
    

    if not force_edit:        
        data = "upcancel {} {} {}".format(message.chat_id,message.id,user_msg.sender_id)
        buts = [KeyboardButtonCallback("cancel upload.",data.encode("UTF-8"))]
        msg = await message.reply("**uploading:** `{}`".format(file_name),buttons=buts)

    else:
        msg = message

    uploader_id = None
    if queue is not None:
        torlog.info(f"waiting for the worker here for {file_name}")
        msg = await message.reply("**uploading:** `{}`".format(file_name),buttons=buts)
        uploader_id = await queue.get()
        torlog.info(f"waiting over for the worker here for {file_name} acquired worker {uploader_id}")

    out_msg = None
    start_time = time.time()
    tout = get_val("EDIT_SLEEP_SECS")
    opath = path
    
    if user_msg is not None:
        dis_thumb = user_db.get_var("DISABLE_THUMBNAIL", user_msg.sender_id)
        if dis_thumb is False or dis_thumb is None:
            thumb_path = user_db.get_thumbnail(user_msg.sender_id)
            if not thumb_path:
                thumb_path = None
    
    try:
        if get_val("FAST_UPLOAD"):
             torlog.info("fast upload is enabled.")
        with open(path,"rb") as filee:
                path = await upload_file(message.client,filee,file_name,
                lambda c,t: progress(c,t,msg,file_name,start_time,tout,message,database)
                )

        if user_msg is not None:
            force_docs = user_db.get_var("FORCE_DOCUMENTS",user_msg.sender_id)  
        else:
            force_docs = None
        
        if force_docs is None:
            force_docs = get_val("FORCE_DOCUMENTS")
    
        if message.media and force_edit:
            out_msg = await msg.edit(
                file=path,
                text=caption_str
            )
        else:
            
            if ftype == "video" and not force_docs:
                try:
                    if thumb_path is not None:
                        thumb = thumb_path
                    else:
                        thumb = await thumb_manage.get_thumbnail(opath)
                except:
                    thumb = None
                    torlog.exception("error in thumbnail.")
                try:
                    attrs, _ = get_attributes(opath,supports_streaming=True)
                    out_msg = await msg.client.send_file(
                        msg.to_id,
                        file=path,
                        parse_mode="html",
                        thumb=thumb,
                        caption=caption_str,
                        reply_to=message.id,
                        supports_streaming=True,
                        progress_callback=lambda c,t: progress(c,t,msg,file_name,start_time,tout,message,database),
                        attributes=attrs
                    )
                except VideoContentTypeInvalidError:
                    attrs, _ = get_attributes(opath,force_document=True)
                    torlog.warning("streamable file send failed, fallback to document.")
                    out_msg = await msg.client.send_file(
                        msg.to_id,
                        file=path,
                        parse_mode="html",
                        caption=caption_str,
                        thumb=thumb,
                        reply_to=message.id,
                        force_document=True,
                        progress_callback=lambda c,t: progress(c,t,msg,file_name,start_time,tout,message,database),
                        attributes=attrs
                    )
                except Exception:
                    torlog.error("error: {}".format(traceback.format_exc()))
            elif ftype == "audio" and not force_docs:
                # not sure about this if
                attrs, _ = get_attributes(opath)
                out_msg = await msg.client.send_file(
                    msg.to_id,
                    file=path,
                    parse_mode="html",
                    caption=caption_str,
                    reply_to=message.id,
                    progress_callback=lambda c,t: progress(c,t,msg,file_name,start_time,tout,message,database),
                    attributes=attrs
                )
            else:
                if force_docs:
                    attrs, _ = get_attributes(opath,force_document=True)
                    out_msg = await msg.client.send_file(
                        msg.to_id,
                        file=path,
                        parse_mode="html",
                        caption=caption_str,
                        reply_to=message.id,
                        force_document=True,
                        progress_callback=lambda c,t: progress(c,t,msg,file_name,start_time,tout,message,database),
                        attributes=attrs,
                        thumb=thumb_path
                    )
                else:
                    attrs, _ = get_attributes(opath)
                    out_msg = await msg.client.send_file(
                        msg.to_id,
                        file=path,
                        parse_mode="html",
                        caption=caption_str,
                        reply_to=message.id,
                        progress_callback=lambda c,t: progress(c,t,msg,file_name,start_time,tout,message,database),
                        attributes=attrs,
                        thumb=thumb_path
                    )
    except Exception as e:
        if str(e).find("cancel") != -1:
            torlog.info("canceled upload")
            await msg.edit(f"failed to upload {e}")
        else:
            torlog.exception("in telegram upload")
            await msg.edit(f"failed to upload {e}")
    finally:
        if queue is not None:
            await queue.put(uploader_id)
            torlog.info(f"freed uploader with id {uploader_id}")
                

    if out_msg is None:
        return None
    if out_msg.id != msg.id:
        await msg.delete()
    
    return out_msg


def black_list_exts(file):
    for i in ['!qb']:
        if str(file).lower().endswith(i):
            return True
    
    return False

#async def upload_single_file(message, local_file_name, caption_str, from_user, edit_media):
async def upload_single_file(path, message, force_edit,database=None,thumb_image_path=None,user_msg=None):
    if database is not None:
        if database.get_cancel_status(message.chat_id,message.id):
            # add os remove here
            return None
    if not os.path.exists(path):
        return None
    
    queue = message.client.exqueue

    file_name = os.path.basename(path)
    caption_str = ""
    caption_str += "<code>"
    caption_str += file_name
    caption_str += "</code>"

    if user_msg is None:
        user_msg = await message.get_reply_message()

    force_docs = get_val("FORCE_DOCUMENTS")
    if user_msg is not None:
        force_docs = user_db.get_var("FORCE_DOCUMENTS",user_msg.sender_id)
    
    # Avoid Flood in Express
    await asyncio.sleep(int(get_val("EDIT_SLEEP_SECS"))+5)

    metadata = extractMetadata(createParser(path))
    
    if metadata is not None:
        # handle none for unknown
        metadata = metadata.exportDictionary()
        try:
            mime = metadata.get("Common").get("MIME type")
        except:
            mime = metadata.get("Metadata").get("MIME type")

        ftype = mime.split("/")[0]
        ftype = ftype.lower().strip()
    else:
        ftype = "unknown"

    thonmsg = message
    message = await message.client.pyro.get_messages(message.chat_id, message.id)
    tout = get_val("EDIT_SLEEP_SECS")
    sent_message = None
    start_time = time.time()
    #
    if user_msg is not None:
        dis_thumb = user_db.get_var("DISABLE_THUMBNAIL", user_msg.sender_id)
        if dis_thumb is False or dis_thumb is None:
            thumb_image_path = user_db.get_thumbnail(user_msg.sender_id)
            if not thumb_image_path:
                thumb_image_path = None
    #
    uploader_id = None
    try:
        message_for_progress_display = message
        if not force_edit:
            data = "upcancel {} {} {}".format(message.chat.id,message.message_id,user_msg.sender_id)
            markup = InlineKeyboardMarkup([[InlineKeyboardButton("cancel upload.", callback_data=data.encode("UTF-8"))]])
            message_for_progress_display = await message.reply_text(
                "**starting upload of** `{}`".format(os.path.basename(path)),
                reply_markup=markup
            )

            if queue is not None:
                torlog.info(f"waiting for the worker here for {file_name}")
                message_for_progress_display = await message_for_progress_display.edit(f"{message_for_progress_display.text}\nwaiting in uploading queue.. ")
                uploader_id = await queue.get()
                torlog.info(f"waiting over for the worker here for {file_name} acquired worker {uploader_id}")
        
        if ftype == "video" and not force_docs:
            metadata = extractMetadata(createParser(path))
            duration = 0
            if metadata.has("duration"):
                duration = metadata.get('duration').seconds
            #
            width = 1280
            height = 720
            if thumb_image_path is None:
                thumb_image_path = await thumb_manage.get_thumbnail(path)
                # get the correct width, height, and duration for videos greater than 10MB

            thumb = None
            if thumb_image_path is not None and os.path.isfile(thumb_image_path):
                thumb = thumb_image_path
            
            # send video
            if force_edit and message.photo:
                sent_message = await message.edit_media(
                    media=InputMediaVideo(
                        media=path,
                        thumb=thumb,
                        parse_mode="html",
                        width=width,
                        height=height,
                        duration=duration,
                        supports_streaming=True,
                        caption=caption_str
                    )
                    # quote=True,
                )
            else:
                sent_message = await message.reply_video(
                    video=path,
                    # quote=True,
                    parse_mode="html",
                    duration=duration,
                    width=width,
                    height=height,
                    thumb=thumb,
                    caption=caption_str,
                    supports_streaming=True,
                    disable_notification=True,
                    # reply_to_message_id=message.reply_to_message.message_id,
                    progress=progress_for_pyrogram,
                    progress_args=(
                        f"uploading {os.path.basename(path)}",
                        message_for_progress_display,
                        start_time,
                        tout,
                        thonmsg.client.pyro,
                        message,
                        database,
                        markup
                    )
                )
            if thumb is not None:
                os.remove(thumb)
        elif ftype == "audio" and not force_docs:
            metadata = extractMetadata(createParser(path))
            duration = 0
            title = ""
            artist = ""
            if metadata.has("duration"):
                duration = metadata.get('duration').seconds
            if metadata.has("title"):
                title = metadata.get("title")
            if metadata.has("artist"):
                artist = metadata.get("artist")
            
            thumb = None
            if thumb_image_path is not None and os.path.isfile(thumb_image_path):
                thumb = thumb_image_path
            # send audio
            if force_edit and message.photo:
                sent_message = await message.edit_media(
                    media=InputMediaAudio(
                        media=path,
                        thumb=thumb,
                        parse_mode="html",
                        duration=duration,
                        performer=artist,
                        title=title,
                        caption=caption_str
                    )
                    # quote=True,
                )
            else:
                sent_message = await message.reply_audio(
                    audio=path,
                    # quote=True,
                    parse_mode="html",
                    duration=duration,
                    performer=artist,
                    title=title,
                    caption=caption_str,
                    thumb=thumb,
                    disable_notification=True,
                    # reply_to_message_id=message.reply_to_message.message_id,
                    progress=progress_for_pyrogram,
                    progress_args=(
                        f"uploading {os.path.basename(path)}",
                        message_for_progress_display,
                        start_time,
                        tout,
                        thonmsg.client.pyro,
                        message,
                        database,
                        markup
                    )
                )
            if thumb is not None:
                os.remove(thumb)
        else:
            # if a file, don't upload "thumb"
            # this "diff" is a major derp -_- 😔😭😭
            thumb = None
            if thumb_image_path is not None and os.path.isfile(thumb_image_path):
                thumb = thumb_image_path
            #
            # send document
            if force_edit and message.photo:
                sent_message = await message.edit_media(
                    media=InputMediaDocument(
                        media=path,
                        caption=caption_str,
                        thumb=thumb,
                        parse_mode="html"
                    )
                    # quote=True,
                )
            else:
                sent_message = await message.reply_document(
                    document=path,
                    # quote=True,
                    thumb=thumb,
                    parse_mode="html",
                    disable_notification=True,
                    # reply_to_message_id=message.reply_to_message.message_id,
                    progress=progress_for_pyrogram,
                    caption=caption_str,
                    progress_args=(
                        f"💠 Uploading {os.path.basename(path)}",
                        message_for_progress_display,
                        start_time,
                        tout,
                        thonmsg.client.pyro,
                        message,
                        database,
                        markup
                    )
                )
            if thumb is not None:
                os.remove(thumb)
    except Exception as e:
        if str(e).find("cancel") != -1:
            torlog.info("canceled upload")
            try:
                await message_for_progress_display.edit(f"failed to upload {e}")
            except:pass
        else:
            try:
                await message_for_progress_display.edit(f"failed to upload {e}")
            except:pass
            torlog.exception("in pyrogram upload")
    else:
        if message.message_id != message_for_progress_display.message_id:
            await message_for_progress_display.delete()
    finally:
        if queue is not None and uploader_id is not None:
            await queue.put(uploader_id)
            torlog.info(f"freed uploader with id {uploader_id}")
    #os.remove(path)
    if sent_message is None:
        return None
    sent_message = await thonmsg.client.get_messages(sent_message.chat.id, ids=sent_message.message_id)
    return sent_message



# -*- coding: utf-8 -*-

import asyncio,shlex,logging,time,os,aiohttp,shutil
import orjson as json
from telethon.hints import MessageLike
from telethon.tl.types import KeyboardButtonCallback, KeyboardButtonUrl
from typing import Union,List,Tuple,Dict,Optional
from ..functions.Human_Format import human_readable_bytes
from ..functions.tele_upload import upload_handel
from ..core.getVars import get_val
from PIL import Image

torlog = logging.getLogger(__name__)

# attempt to decorate error prone areas
import traceback
def skipTorExp(func):
    def wrap_func(*args,**kwargs):
        try:
            return func(*args,**kwargs)
        except Exception as e:
            torlog.error(e)
            return
    return wrap_func

async def cli_call(cmd: Union[str,List[str]]) -> Tuple[str,str]:
    if isinstance(cmd,str):
        cmd = shlex.split(cmd)
    elif isinstance(cmd,(list,tuple)):
        pass
    else:
        return None,None
    
    process = await asyncio.create_subprocess_exec(
        *cmd,
        stderr=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE
    )

    stdout, stderr = await process.communicate()
    
    stdout = stdout.decode().strip()
    stderr = stderr.decode().strip()

    with open("test.txt","w",encoding="UTF-8") as f:
        f.write(stdout)
    
    return stdout, stderr


async def get_yt_link_details(url: str) -> Union[Dict[str,str], None]:
    cmd = "youtube-dl --no-warnings --youtube-skip-dash-manifest --dump-json"
    cmd = shlex.split(cmd)
    if "hotstar" in url:
        cmd.append("--geo-bypass-country")
        cmd.append("IN")
    cmd.append(url)
    
    out, error = await cli_call(cmd)
    if error:
        torlog.error(f"error occured: {error} for url {url}")
    # sanitize the json
    out = out.replace("\n",",")
    out = "[" + out + "]"
    try:
        return json.loads(out)[0], None
    except:
        torlog.exception("an error occured while parsing the json.\n")
        return None, error

async def get_max_thumb(data: dict, suid: str) -> str:
    thumbnail = data.get("thumbnail")
    thumb_path = None

    # alot of context management XD
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(thumbnail) as resp:
                thumb_path = os.path.join(os.getcwd(),"userdata")
                if not os.path.exists(thumb_path):
                    os.mkdir(thumb_path)

                thumb_path = os.path.join(thumb_path,f"{suid}.webp")
                with open(thumb_path,"wb") as ifile:
                    ifile.write(await resp.read())

        Image.open(thumb_path).convert("RGB").save(thumb_path)

        return thumb_path
    except:
        torlog.exception("error in thumbnail generation.")
        return None

async def create_quality_menu(url: str,message: MessageLike, message1: MessageLike,jsons: Optional[str] = None, suid: Optional[str] = None):
    if jsons is None:
        data, err = await get_yt_link_details(url)
        suid = str(time.time()).replace(".","")
    else:
        data = jsons

    with open("test.txt","w") as f:
        f.write(json.dumps(data).decode("UTF-8"))

    if data is None:
        await message.edit("error failed parsing.")
        return None, err
    else:
        unique_formats = dict()
        for i in data.get("formats"):
            c_format = i.get("format_note")
            if c_format is None:
                c_format = i.get("height")
            if not c_format in unique_formats:
                if i.get("filesize") is not None:
                    unique_formats[c_format] = [i.get("filesize"),i.get("filesize")]
                else:
                    unique_formats[c_format] = [0,0]

            else:
                if i.get("filesize") is not None:
                    if unique_formats[c_format][0] > i.get("filesize"):
                        unique_formats[c_format][0] = i.get("filesize")
                    else:
                        unique_formats[c_format][1] = i.get("filesize")

        buttons = list()
        for i in unique_formats.keys():
            
            # add human bytes here
            if i == "tiny":
                text = f"tiny [{human_readable_bytes(unique_formats[i][0])} - {human_readable_bytes(unique_formats[i][1])}] ➡️"
                cdata = f"ytdlsmenu|{i}|{message1.sender_id}|{suid}" # add user id
            else:
                text = f"{i} [{human_readable_bytes(unique_formats[i][0])} - {human_readable_bytes(unique_formats[i][1])}] ➡️"
                cdata = f"ytdlsmenu|{i}|{message1.sender_id}|{suid}" # add user id
            buttons.append([KeyboardButtonCallback(text,cdata.encode("UTF-8"))])
        buttons.append([KeyboardButtonCallback("Audios ➡️",f"ytdlsmenu|audios|{message1.sender_id}|{suid}")])
        await message.edit("Choose a quality/option available below.",buttons=buttons)
        
        if jsons is None:
            path = os.path.join(os.getcwd(),'userdata')
            
            if not os.path.exists(path):
                os.mkdir(path)
            
            path = os.path.join(path,f"{suid}.json")
            
            with open(path,"w",encoding="UTF-8") as file:
                file.write(json.dumps(data).decode("UTF-8"))



    return True,None
        
async def handle_ytdl_command(e: MessageLike):
    if not e.is_reply:
        await e.reply("reply to a youtube video link.")
        return
    msg = await e.get_reply_message()
    msg1 = await e.reply("processing the given link.....")
    if msg.text.find("http") != -1:
        res, err = await create_quality_menu(msg.text.strip(),msg1,msg)
        if res is None:
            await msg1.edit(f"<code>invalid link provided.\n{err}</code>",parse_mode="html")
    else:
        await e.reply("invalid link provided.")

async def handle_ytdl_callbacks(e: MessageLike):
    data = e.data.decode("UTF-8")
    data = data.split("|")
    
    if data[0] == "ytdlsmenu":
        if data[2] != str(e.sender_id):
            await e.answer("this isn't yours.")
            return
        
        path = os.path.join(os.getcwd(),'userdata',data[3]+".json")
        if os.path.exists(path):
            with open(path) as file:
                ytdata = json.loads(file.read())
                buttons = list()
                if data[1] == "audios":
                    for i in ["64K","128K","320K"]:
                        text = f"{i} [MP3]"
                        cdata = f"ytdldfile|{i}|{e.sender_id}|{data[3]}"
                        buttons.append([KeyboardButtonCallback(text,cdata.encode("UTF-8"))])
                else:
                    j = 0
                    for i in ytdata.get("formats"):
                        c_format = i.get("format_note")
                        format_id = i.get('format_id')
                        height = i.get('format')
                        if c_format is None:
                            c_format = str(i.get("height"))
                            format_id = f"xxother{j}"
                            height = i.get('format')
                        if not c_format == data[1]:
                            continue
                        
                        
                        if not height:
                            continue
                            
                        text = f"{height} [{i.get('ext')}] [{human_readable_bytes(i.get('filesize'))}] {str(i.get('vcodec'))}"
                        cdata = f"ytdldfile|{format_id}|{e.sender_id}|{data[3]}"
                        
                        buttons.append([KeyboardButtonCallback(text,cdata.encode("UTF-8"))])
                        j+=1
                
                buttons.append([KeyboardButtonCallback("Go Back",f"ytdlmmenu|{data[2]}|{data[3]}")])
                await e.edit(f"files for quality {data[1]}, at the end it is the video codec. mostly prefer the last one with you desired extension if you want streamable video.",buttons=buttons)
                


        else:
            await e.answer("try again, something went wrong.",alert=True)
            await e.delete()
    elif data[0] == "ytdlmmenu":
        if data[1] != str(e.sender_id):
            await e.answer("this isn't yours.")
            return
        path = os.path.join(os.getcwd(),'userdata',data[2]+".json")
        if os.path.exists(path):
            with open(path,encoding="UTF-8") as file:
                ytdata = json.loads(file.read())
                await create_quality_menu("",await e.get_message(),e,ytdata,data[2])

        else:
            await e.answer("try again, something went wrong.",alert=True)
            await e.delete()

async def handle_ytdl_file_download(e: MessageLike):
    # ytdldfile | format_id | sender_id | suid | is_audio

    data = e.data.decode("UTF-8")
    data = data.split("|")
    
    
    if data[2] != str(e.sender_id):
        await e.answer("this isn't yours.")
        return
    else:
        await e.answer("crunching data.....")
    
    await e.edit(buttons=None)

    is_audio = False

    path = os.path.join(os.getcwd(),'userdata',data[3]+".json")
    if os.path.exists(path):
        with open(path,encoding="UTF-8") as file:
            ytdata = json.loads(file.read())
            yt_url = ytdata.get("webpage_url")
            thumb_path = await get_max_thumb(ytdata,data[3])

            op_dir = os.path.join(os.getcwd(),'userdata',data[3])
            if not os.path.exists(op_dir):
                os.mkdir(op_dir)
            if data[1].startswith("xxother"):
                data[1] = data[1].replace("xxother","")
                data[1] = int(data[1])
                j = 0
                for i in ytdata.get("formats"):
                    if j == data[1]:
                        data[1] = i.get("format_id")
                    j +=1
            else:
                for i in ytdata.get("formats"):
                    if i.get("format_id") == data[1]:
                        print(i)
                        if i.get("acodec") is not None:
                            if "none" not in i.get("acodec"):
                                is_audio = True
                            
                    
            if data[1].endswith("K"):
                cmd = f"youtube-dl -i --extract-audio --add-metadata --audio-format mp3 --audio-quality {data[1]} -o '{op_dir}/%(title)s.%(ext)s' {yt_url}"

            else:
                if is_audio:
                    cmd = f"youtube-dl --continue --embed-subs --no-warnings --hls-prefer-ffmpeg --prefer-ffmpeg -f {data[1]} -o {op_dir}/%(title)s.%(ext)s {yt_url}"
                else:
                    cmd = f"youtube-dl --continue --embed-subs --no-warnings --hls-prefer-ffmpeg --prefer-ffmpeg -f {data[1]}+bestaudio[ext=m4a]/best -o {op_dir}/%(title)s.%(ext)s {yt_url}"
            
            out, err = await cli_call(cmd)
            
            if not err:
                
                # TODO Fix the original thumbnail
                # rdict = await upload_handel(op_dir,await e.get_message(),e.sender_id,dict(),thumb_path=thumb_path)
                
                rdict = await upload_handel(op_dir,await e.get_message(),e.sender_id,dict(), user_msg=e)
                await print_files(e,rdict)
                
                shutil.rmtree(op_dir)
                os.remove(thumb_path)
                os.remove(path)
            else:
                torlog.error(err)
                omess = await e.get_message()
                omess1 = await omess.get_reply_message()
                if "HTTP Error 429" in err:
                    emsg = "http error 429: too many requests, try after a while."
                else:
                    emsg = "an error has occured. trying to upload any files that are found here."
                await omess.edit(emsg)
                if omess1 is None:
                    await omess.respond(emsg)
                else:
                    await omess1.reply(emsg)
                
                rdict = await upload_handel(op_dir,await e.get_message(),e.sender_id,dict(), user_msg=e)
                await print_files(e,rdict)

                try:
                    shutil.rmtree(op_dir)
                    os.remove(thumb_path)
                    os.remove(path)
                except:
                    pass

    else:
        await e.delete()
        await e.answer("try again, something went wrong.",alert=True)
        await e.delete()

async def handle_ytdl_playlist(e: MessageLike) -> None:
    if not e.is_reply:
        await e.reply("reply to a youtube playlist link.")
        return
    url = await e.get_reply_message()
    url = url.text.strip()
    cmd = f"youtube-dl -i --flat-playlist --dump-single-json {url}"
    
    msg = await e.reply("processing your youtube playlist download request..")

    # cancel the playlist if time exceed 5 mins
    try:
        out, err = await asyncio.wait_for(cli_call(cmd),300)
    except asyncio.TimeoutError:
        await msg.edit("processing time exceeded... the playlist seem to long to be worked with\n if the playlist is short and you think its error report back.")
        return
    
    if err:
        await msg.edit(f"failed to load the playlist with the error: <code>{err}</code>",parse_mode="html")
        return
    

    try:
        pldata = json.loads(out)
        entities = pldata.get("entries")
        if len(entities) <= 0:
            await msg.edit("cannot load the videos from this playlist, ensure that the playlist is not <code>'My Mix or Mix'</code>. It should be a public or unlisted youtube playlist.")
            return

        entlen = len(entities)
        keybr = list()
        
        # limit the max vids
        if entlen > get_val("MAX_YTPLAYLIST_SIZE"):

            await msg.edit(f"playlist too large max {get_val('MAX_YTPLAYLIST_SIZE')} vids allowed as of now. This has {entlen}")
            return


        # format> ytdlplaylist | quality | suid | sender_id
        suid = str(time.time()).replace(".","")

        for i in ["144","240","360","480","720","1080","1440","2160"]:
            keybr.append([KeyboardButtonCallback(text=f"{i}p All videos",data=f"ytdlplaylist|{i}|{suid}|{e.sender_id}")])

        keybr.append([KeyboardButtonCallback(text=f"Best All videos",data=f"ytdlplaylist|best|{suid}|{e.sender_id}")])
        
        
        keybr.append([KeyboardButtonCallback(text="best all audio only. [340k]",data=f"ytdlplaylist|320k|{suid}|{e.sender_id}")])
        keybr.append([KeyboardButtonCallback(text="medium all audio only. [128k]",data=f"ytdlplaylist|128k|{suid}|{e.sender_id}")])
        keybr.append([KeyboardButtonCallback(text="worst all audio only. [64k]",data=f"ytdlplaylist|64k|{suid}|{e.sender_id}")])

        await msg.edit(f"found {entlen} videos in the playlist.",buttons=keybr) 

        path = os.path.join(os.getcwd(),'userdata')
        
        if not os.path.exists(path):
            os.mkdir(path)
        
        path = os.path.join(path,f"{suid}.json")
        
        with open(path,"w",encoding="UTF-8") as file:
            file.write(json.dumps(pldata).decode("UTF-8"))

    except:
        await msg.edit("failed to parse the playlist. check logs for any error messages.")
        torlog.exception("playlist parse failed") 

async def handle_ytdl_playlist_down(e: MessageLike) -> None:
    # ytdlplaylist | quality | suid | sender_id
    
    data = e.data.decode("UTF-8").split("|")
    
    if data[3] != str(e.sender_id):
        await e.answer("this isn't yours.")
        return
    else:
        await e.answer("crunching data.....")

    await e.edit(buttons=None)
    path = os.path.join(os.getcwd(),"userdata",data[2]+".json")
    if os.path.exists(path):
        await e.answer("processing, please wait.")
        opdir = os.path.join(os.getcwd(),"userdata",data[2])
        if not os.path.exists(opdir):
            os.mkdir(opdir)

        with open(path) as file:
            pldata = json.loads(file.read())
        url = pldata.get("webpage_url")

        if data[1].endswith("k"):
            audcmd = f"youtube-dl -i --extract-audio --add-metadata --audio-format mp3 --audio-quality {data[1]} -o '{opdir}/%(playlist_index)s - %(title)s.%(ext)s' {url}"
            out, err = await cli_call(audcmd)
            if err:
                await e.reply(f"failed to download the audios <code>{err}</code>",parse_mode="html")
            else:
                rdict = await upload_handel(opdir, await e.get_message(), e.sender_id, dict(), user_msg=e)
                await print_files(e,rdict)
        else:
            if data[1] == "best":
                vidcmd = f"youtube-dl --continue --embed-subs --no-warnings --prefer-ffmpeg -f 'bestvideo[ext=mp4]+bestaudio[ext=m4a]/best' -o '{opdir}/%(playlist_index)s - %(title)s.%(ext)s' {url}"
            else:
                vidcmd = f"youtube-dl --continue --embed-subs --no-warnings --prefer-ffmpeg -f 'bestvideo[ext=mp4,height<={data[1]}]+bestaudio[ext=m4a]/best' -o '{opdir}/%(playlist_index)s - %(title)s.%(ext)s' {url}"
            out, err = await cli_call(vidcmd)
            if err:
                await e.reply(f"failed to download the videos <code>{err}</code>",parse_mode="html")
            else:
                rdict = await upload_handel(opdir, await e.get_message(), e.sender_id, dict(), user_msg=e)
                await print_files(e,rdict)
        shutil.rmtree(opdir)
        os.remove(path)
    else:
        await e.delete()
        await e.answer("something went wrong, try again.",alert=True)
        torlog.error("the file for that suid was not found.")

async def print_files(e,files):
    
    msg = "uploaded.\n"
    if len(files) == 0:
        return
    
    chat_id = e.chat_id

    for i in files.keys():
        link = f'https://t.me/c/{str(chat_id)[4:]}/{files[i]}'
        msg += f'<a href="{link}">{i}</a>\n'
     
    rmsg = await e.client.get_messages(e.chat_id,ids=e.message_id)
    rmsg = await rmsg.get_reply_message()
    if rmsg is None:
        #msg += "\n<a href='tg://user?id={}'>Done<a>".format(rmsg.sender_id)
        msg += "\n<a href='tg://user?id={}'>finished leech.</a>".format(e.sender_id)
        await e.reply(msg,parse_mode="html")
    else:
        msg += "\n<a href='tg://user?id={}'>Done</a>".format(rmsg.sender_id)
        await rmsg.reply(msg,parse_mode="html")

    if len(files) < 2:
        return

    ids = list()
    for i in files.keys():
        ids.append(files[i])
    
    msgs = await e.client.get_messages(e.chat_id,ids=ids)
    for i in msgs:
        index = None
        for j in range(0,len(msgs)):
            index = j
            if ids[j] == i.id:
                break
        nextt,prev = "",""
        chat_id = str(e.chat_id)[4:]
        buttons = []
        if index == 0:
            nextt = f'https://t.me/c/{chat_id}/{ids[index+1]}'
            buttons.append(
                KeyboardButtonUrl("Next", nextt)
            )
            nextt = f'<a href="{nextt}">Next</a>\n'
        elif index == len(msgs)-1:
            prev = f'https://t.me/c/{chat_id}/{ids[index-1]}'
            buttons.append(
                KeyboardButtonUrl("Prev", prev)
            )
            prev = f'<a href="{prev}">Prev</a>\n'
        else:
            nextt = f'https://t.me/c/{chat_id}/{ids[index+1]}'
            buttons.append(
                KeyboardButtonUrl("Next", nextt)
            )
            nextt = f'<a href="{nextt}">Next</a>\n'
            
            prev = f'https://t.me/c/{chat_id}/{ids[index-1]}'
            buttons.append(
                KeyboardButtonUrl("Prev", prev)
            )
            prev = f'<a href="{prev}">Prev</a>\n'
            
        try:
            #await i.edit("{} {} {}".format(prev,i.text,nextt),parse_mode="html")
            await i.edit(buttons=buttons)
        except:pass
        await asyncio.sleep(1)

#todo
# Add the YT playlist feature here
# Add the YT channels feature here 


import setuptools

setuptools.setup(
    name="tgtk",
    version="1.1.6",
    author="reaitten",
    description="A Telegram Bot written in Python to mirror files on the internet onto Telegram",
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/reaitten/tgtk",
    project_urls={
        "Bug Tracker": "https://github.com/reaitten/tgtk/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 5 - Production/Stable"
    ],
    packages=setuptools.find_packages(),
    install_requires=open('requirements.txt', 'r', encoding='utf-8').read().split('\n'),
    scripts=['start.sh'],
    python_requires=">=3.8",
)



# -*- coding: utf-8 -*-


import psycopg2,os,datetime
import psycopg2.extras
from ..functions.pg_plugin import DataBaseHandle
from ..consts.ExecVarsSample import ExecVars
import json
#this will handel the transaction with completed torrents

# this now will handle the setting for the bot
class tkdb(DataBaseHandle):
    def __init__(self,dburl=None):
        # *** QUERIES ***
        if dburl is None:
            dburl = os.environ.get("DB_URI",None)
            if dburl is None:
                dburl = ExecVars.DB_URI

        super().__init__(dburl)

        settings_schema = """CREATE TABLE IF NOT EXISTS ttk_config (
            id SERIAL PRIMARY KEY NOT NULL,
            var_name VARCHAR(50) NOT NULL UNIQUE,
            var_value VARCHAR(2000) DEFAULT NULL,
            vtype VARCHAR(20) DEFAULT NULL,
            blob_val BYTEA DEFAULT NULL,
            date_changed TIMESTAMP NOT NULL
        )"""

        cur = self.scur()
        try:
            # Sometimes multiple instance try to creat which may cause this error
            cur.execute(settings_schema)
        except psycopg2.errors.UniqueViolation: # pylint: disable=no-member
            pass
        
        self._conn.commit()
        self.ccur(cur)

    def set_variable(self,var_name,var_value,update_blob=False,blob_val=None):
        #todo implement blob - done
        # remember to handle the memoryview
        vtype = "str"
        if isinstance(var_value,bool):
            vtype = "bool"
        elif isinstance(var_value,int):
            vtype = "int"
        
        if update_blob:
            vtype = "blob"

        sql = "SELECT * FROM ttk_config WHERE var_name=%s"
        cur = self.scur()
        
        cur.execute(sql,(var_name,))
        if cur.rowcount > 0:
            if not update_blob:
                sql = "UPDATE ttk_config SET var_value=%s , vtype=%s WHERE var_name=%s"
            else:
                sql = "UPDATE ttk_config SET blob_val=%s , vtype=%s WHERE var_name=%s"
                var_value = blob_val

            cur.execute(sql,(var_value,vtype,var_name))
        else:
            if not update_blob:
                sql = "INSERT INTO ttk_config(var_name,var_value,date_changed,vtype) VALUES(%s,%s,%s,%s)"
            else:
                sql = "INSERT INTO ttk_config(var_name,blob_val,date_changed,vtype) VALUES(%s,%s,%s,%s)"
                var_value = blob_val
            
            cur.execute(sql,(var_name,var_value,datetime.datetime.now(),vtype))

        self.ccur(cur)
    
    def get_variable(self,var_name):
        sql = "SELECT * FROM ttk_config WHERE var_name=%s"
        cur = self.scur()
        
        cur.execute(sql,(var_name,))
        if cur.rowcount > 0:
            row = cur.fetchone()
            vtype = row[3]
            val = row[2]
            if vtype == "int":
                val = int(row[2])
            elif vtype == "str":
                val = str(row[2])
            elif vtype == "bool":
                if row[2] == "true":
                    val = True
                else:
                    val = False

            return val,row[4]
        else:
            return None,None

        self.ccur(cur)
        

    def __del__(self):
        super().__del__()

class tkupload(DataBaseHandle):
    # Common across all the objs
    cache_store = {}

    def __init__(self,dburl=None):
        # *** QUERIES ***
        if dburl is None:
            dburl = os.environ.get("DB_URI",None)
            if dburl is None:
                dburl = ExecVars.DB_URI

        super().__init__(dburl)

        uploads_schema = """CREATE TABLE IF NOT EXISTS ttk_uploads (
            id SERIAL PRIMARY KEY NOT NULL,
            chat_id VARCHAR(30), --Moslty Overkill
            message_id VARCHAR(30), --Moslty Overkill
            cancel BOOLEAN DEFAULT false,
            is_batch BOOLEAN DEFAULT false
        )"""

        cur = self.scur()
        
        try:
            # Sometimes multiple instance try to creat which may cause this error
            cur.execute(uploads_schema)
        except psycopg2.errors.UniqueViolation: # pylint: disable=no-member
            pass

        self._conn.commit()
        self.ccur(cur)
        

    def register_upload(self,chat_id,mes_id,is_batch=False):
        chat_id = str(chat_id)
        mes_id = str(mes_id)

        cur = self.scur()
        sql = "SELECT * FROM ttk_uploads WHERE chat_id=%s and message_id=%s"
        
        cur.execute(sql,(chat_id,mes_id))

        if cur.rowcount > 0:
            row = cur.fetchone()
            sql = "DELETE FROM ttk_uploads WHERE id=%s"
            cur.execute(sql,(row[0],))

        sql = "INSERT INTO ttk_uploads(chat_id,message_id,is_batch) VALUES(%s,%s,%s)"
        
        cur.execute(sql,(chat_id,mes_id,is_batch))
        self.cache_store["{}-{}".format(chat_id, mes_id)] = False
        self.ccur(cur)

    def cancel_download(self,chat_id,mes_id):
        chat_id = str(chat_id)
        mes_id = str(mes_id)

        cur = self.scur()
        sql = "SELECT * FROM ttk_uploads WHERE chat_id=%s and message_id=%s"
        
        cur.execute(sql,(chat_id,mes_id))
        try:
            self.cache_store["{}-{}".format(chat_id, mes_id)] = True
        except KeyError:
            pass
        if cur.rowcount == 0:
            self.ccur(cur)
            return False
        else:
            sql = "UPDATE ttk_uploads SET cancel=TRUE WHERE chat_id=%s and message_id=%s"
            cur.execute(sql,(chat_id,mes_id))
            self.ccur(cur)
            return True

    def get_cancel_status(self,chat_id,mes_id):
        chat_id = str(chat_id)
        mes_id = str(mes_id)

        cur = self.scur()
        try:
            # Caching in memory
            return self.cache_store["{}-{}".format(chat_id, mes_id)]
        except KeyError:
            pass
        sql = "SELECT * FROM ttk_uploads WHERE chat_id=%s and message_id=%s"
        
        cur.execute(sql,(chat_id,mes_id))

        if cur.rowcount == 0:
            self.ccur(cur)
            return False
        else:
            row = cur.fetchone()
            cancel = row[3]
            self.ccur(cur)
            return cancel


    def deregister_upload(self,chat_id,mes_id):
        chat_id = str(chat_id)
        mes_id = str(mes_id)
        
        try:
            del self.cache_store["{}-{}".format(chat_id, mes_id)]
        except KeyError:
            pass
        sql = "DELETE FROM ttk_uploads WHERE chat_id=%s and message_id=%s"
        cur = self.scur()
        cur.execute(sql,(chat_id,mes_id))
        self.ccur(cur)


    def __del__(self):
        super().__del__()

class tktorrents(DataBaseHandle):
    def __init__(self,dburl=None):
        if dburl is None:
            dburl = os.environ.get("DB_URI",None)
            if dburl is None:
                dburl = ExecVars.DB_URI

        super().__init__(dburl)
        cur = self.scur()
        table = """CREATE TABLE IF NOT EXISTS ttk_torrents(
            id SERIAL PRIMARY KEY NOT NULL,
            hash_id VARCHAR(100) NOT NULL,
            passw VARCHAR(10) NOT NULL,
            enab BOOLEAN DEFAULT true
        )
        """

        try:
            # Sometimes multiple instance try to creat which may cause this error
            cur.execute(table)
        except psycopg2.errors.UniqueViolation: # pylint: disable=no-member
            pass
        
        self.ccur(cur)

    def add_torrent(self,hash_id,passw):
        sql = "SELECT * FROM ttk_torrents WHERE hash_id=%s"
        cur = self.scur()
        cur.execute(sql,(hash_id,))
        if cur.rowcount > 0:
            sql = "UPDATE ttk_torrents SET passw=%s WHERE hash_id=%s"
            cur.execute(sql,(passw,hash_id))
        else:
            sql = "INSERT INTO ttk_torrents(hash_id,passw) VALUES(%s,%s)"
            cur.execute(sql,(hash_id,passw))
        
        self.ccur(cur)

    def disable_torrent(self,hash_id):
        sql = "SELECT * FROM ttk_torrents WHERE hash_id=%s"
        cur = self.scur()
        cur.execute(sql,(hash_id,))
        if cur.rowcount > 0:
            sql = "UPDATE ttk_torrents SET enab=false WHERE hash_id=%s"
            cur.execute(sql,(hash_id,))

        self.ccur(cur)
        
    def get_password(self,hash_id):
        sql = "SELECT * FROM ttk_torrents WHERE hash_id=%s"
        cur = self.scur()
        cur.execute(sql,(hash_id,))
        if cur.rowcount > 0:
            row = cur.fetchone()
            self.ccur(cur)
            return row[2]
        else:
            self.ccur(cur)
            return False

    def purge_all_torrents(self):
        sql = "DELETE FROM ttk_torrents"
        cur = self.scur()
        cur.execute(sql)
        self.ccur(cur)

class userdb(DataBaseHandle):
    shared_users = {}
    def __init__(self,dburl=None):
        if dburl is None:
            dburl = os.environ.get("DB_URI",None)
            if dburl is None:
                dburl = ExecVars.DB_URI

        super().__init__(dburl)
        cur = self.scur()
        table = """CREATE TABLE IF NOT EXISTS ttk_users(
            id SERIAL PRIMARY KEY NOT NULL,
            user_id VARCHAR(50) NOT NULL,
            json_data VARCHAR(1000) NOT NULL, --Keeping it as json so that it flexible to add stuff.
            rclone_file BYTEA DEFAULT NULL,
            thumbnail BYTEA DEFAULT NULL
        )
        """

        try:
            # Sometimes multiple instance try to creat which may cause this error
            cur.execute(table)
        except psycopg2.errors.UniqueViolation: # pylint: disable=no-member
            pass
        
        self.ccur(cur)

    def get_var(self, var, user_id):
        user_id = str(user_id)
        sql = "SELECT * FROM ttk_users WHERE user_id=%s"
        # search the cache
        user = self.shared_users.get(user_id)
        if user is not None:
            return user.get(var)
        else:
            cur = self.scur(dictcur=True)
            
            cur.execute(sql, (user_id,))
            if cur.rowcount > 0:
                user = cur.fetchone()
                jdata = user.get("json_data")
                jdata = json.loads(jdata)
                self.shared_users[user_id] = jdata
                return jdata.get(var)
            else:
                return None
                

            self.ccur(cur)

    def set_var(self, var, value, user_id):
        user_id = str(user_id)
        sql = "SELECT * FROM ttk_users WHERE user_id=%s"
        # search the cache
        cur = self.scur(dictcur=True)

        user = self.shared_users.get(user_id)
        if user is not None:
            self.shared_users[user_id][var] = value
             
        else:
            
            cur.execute(sql, (user_id,))
            if cur.rowcount > 0:
                user = cur.fetchone()
                jdata = user.get("json_data")
                jdata = json.loads(jdata)
                jdata[var] = value
                self.shared_users[user_id] = jdata
            else:
                self.shared_users[user_id] = {var:value}

        cur.execute(sql, (user_id,))
        if cur.rowcount > 0:
            insql = "UPDATE ttk_users SET json_data = %s where user_id=%s"
            cur.execute(insql, ( json.dumps(self.shared_users.get(user_id)), user_id))

        else:
            insql = "INSERT INTO ttk_users(user_id, json_data) VALUES(%s, %s)"
            cur.execute(insql, (user_id, json.dumps(self.shared_users.get(user_id))))
        
        self.ccur(cur)

    def get_rclone(self, user_id):
        user_id = str(user_id)
        sql = "SELECT * FROM ttk_users WHERE user_id=%s"
        cur = self.scur(dictcur=True)

        cur.execute(sql, (user_id,))
        
        if cur.rowcount > 0:
            row = cur.fetchone()
            self.ccur(cur)

            if row["rclone_file"] is None:
                return False
            else:
                path = os.path.join(os.getcwd(), 'userdata')
                if not os.path.exists(path):
                    os.mkdir(path)
                
                path = os.path.join(path, user_id)
                if not os.path.exists(path):
                    os.mkdir(path)
                
                path = os.path.join(path, "rclone.conf")
                with open(path, "wb") as rfile:
                    rfile.write(row["rclone_file"])
                
                return path
        else:
            return False


    def get_thumbnail(self, user_id):
        user_id = str(user_id)
        sql = "SELECT * FROM ttk_users WHERE user_id=%s"
        cur = self.scur(dictcur=True)

        cur.execute(sql, (user_id,))
        
        
        if cur.rowcount > 0:
            row = cur.fetchone()
            self.ccur(cur)
            if row["thumbnail"] is None:
                return False
            else:
                path = os.path.join(os.getcwd(), 'userdata')
                if not os.path.exists(path):
                    os.mkdir(path)
                
                path = os.path.join(path, user_id)
                if not os.path.exists(path):
                    os.mkdir(path)
                
                path = os.path.join(path, "thumbnail.jpg")
                with open(path, "wb") as rfile:
                    rfile.write(row["thumbnail"])
                
                return path
        else:
            return False

    def set_rclone(self, rclonefile, user_id):
        user_id = str(user_id)

        sql = "SELECT * FROM ttk_users WHERE user_id=%s"
        cur = self.scur(dictcur=True)

        cur.execute(sql, (user_id,))
        if cur.rowcount > 0:
            insql = "UPDATE ttk_users SET rclone_file=%s WHERE user_id=%s"
            cur.execute(insql, (rclonefile, user_id))

        else:
            insql = "INSERT INTO ttk_users(user_id, json_data, rclone_file) VALUES(%s, '{}', %s)"
            cur.execute(insql, (user_id, rclonefile))

        self.ccur(cur)
        return True

    def set_thumbnail(self, thumbnail, user_id):
        user_id = str(user_id)

        sql = "SELECT * FROM ttk_users WHERE user_id=%s"
        cur = self.scur(dictcur=True)

        cur.execute(sql, (user_id,))
        if cur.rowcount > 0:
            insql = "UPDATE ttk_users SET thumbnail=%s WHERE user_id=%s"
            cur.execute(insql, (thumbnail, user_id))

        else:
            insql = "INSERT INTO ttk_users(user_id, json_data, thumbnail) VALUES(%s, '{}', %s)"
            cur.execute(insql, (user_id, thumbnail))

        self.ccur(cur)
        return True


# -*- coding: utf-8 -*-


from ..consts.ExecVarsSample import ExecVars
#from ..core.database_handle import tkdb
from tgtk import SessionVars
import os

def get_val(variable):
    return SessionVars.get_var(variable)

